/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS Channel API
 */
#ifndef _NODEOS_CHANNEL_H_
#define _NODEOS_CHANNEL_H_

#include <nodeos/resource.h>
#include <nodeos/demuxkey.h>
#include <nodeos/packet.h>

typedef struct _an_chan {
	int stuff;
} *an_chan_t;

/*
 * Channel protocol processing information.
 * Currently implemented as a string; e.g. "ETH/IP".
 */
typedef struct {
	int len;
	char data[NODEOS_CHAN_PROTOSIZE-sizeof(int)];
} an_chanproto_t;

/*
 * Channel peer address information.
 * Channel-type specific.
 */
typedef union {
	an_ethaddr_t eth;
	an_sockaddr_t socket;
	an_clockaddr_t clock;
	char data[NODEOS_CHAN_ADDRSIZE];
} an_chanaddr_t;

/*
 * Called when a packet arrives on a channel.
 * The "void *" is an additional argument supplied by the user
 * at channel creation time.
 */
typedef void (*an_chanrecvfunc_t)(an_packet_t, void *);

/* XXX compat */
#ifdef NODEOS_SPEC_COMPAT
#define an_Channel an_chan_t
#define an_ChannelAddress an_chanaddr_t
#define an_ChannelProtocol an_chanproto_t
#define an_DeliverFunc an_chanrecvfunc_t
#define an_inchannelCreate an_inchan_create
#define an_outchannelCreate an_outchan_create
#define an_cutchannelCreate an_cutchan_create
#define an_channelDestroy an_chan_destroy
#define an_channelSend an_chan_send
#define an_channelAddMem an_chan_addmem
#endif

/*
 * Create an in-channel as described by the given parameters.  Every packet
 * matching the given protocol, addressing and demultiplex criteria is
 * handed off to the specified delivery function.  Returns a pointer to the
 * new channel object or zero on an error.
 *
 * ``flow'' is the flow with which the channel should be associated,
 * ``key'' contains the filtering criteria to be applied to every packet,
 * ``proto'' describes the protocol processing to apply to every packet,
 * ``addr'' contains the addressing info appropriate for the protocols,
 * ``mem'' specifies the buffering resources to use for packets,
 * ``cpu'' specifies the CPU resources to use,
 * ``func'' is the packet delivery function,
 * ``farg'' is an additional argument provided to the delivery function.
 */
an_chan_t an_inchan_create(an_flow_t flow, an_demuxkey_t key,
			   an_chanproto_t *proto, an_chanaddr_t *addr,
			   an_memspec_t *mem, an_cpuspec_t *cpu,
			   an_chanrecvfunc_t func, void *farg);

/*
 * Create an out-channel as described by the given parameters.  Every packet
 * sent on the channel undergoes the specified protocol processing and is
 * routed based on the addressing info.  Returns a pointer to the new channel
 * object or zero on an error.
 *
 * ``flow'' is the flow with which the channel should be associated,
 * ``proto'' describes the protocol processing to apply to every packet,
 * ``addr'' contains the addressing info appropriate for the protocols,
 * ``net'' specifies the network resources to use,
 */
an_chan_t an_outchan_create(an_flow_t flow, an_chanproto_t *proto,
			    an_chanaddr_t *addr, an_netspec_t *net);

/*
 * Create a cut-channel by composing the provided in- and out- channels.
 * Additional protocol processing is specified in the given protocol spec.
 * The in-chan and out-chan can no longer be used once they are composed
 * into a cut-channel.
 *
 * ``flow'' is the flow with which the channel should be associated,
 * ``proto'' describes the protocol processing to apply to every packet,
 * ``inchan'' is the in-channel created for receiving packets,
 * ``outchan'' is the out-channel created for sending the packets
 *
 * XXX not implemented.
 */
an_chan_t an_cutchan_create(an_flow_t flow, an_chanproto_t *proto,
			    an_chan_t inchan, an_chan_t outchan);

/*
 * Cleanly shutdown a channel.
 * Further operations on the channel should return an error.  For
 * in-channels, no further packets will be received, though processing
 * of currently active packets continues.
 *
 * ``chan'' is the channel to destroy.
 */
void an_chan_destroy(an_chan_t chan);

#define an_inchan_destroy an_chan_destroy
#define an_outchan_destroy an_chan_destroy
#define an_cutchan_destroy an_chan_destroy

/*
 * Send the indicated packet to an out-channel.
 * Once this function returns, the packet may be safely destroyed or
 * returned to the channel buffer pool.  No indication is provided as
 * to whether the packet was successfully transmitted.  Returns an error
 * if the channel has been destroyed.
 *
 * ``chan'' is the channel to send on.
 * ``packet'' is the packet to send.
 */
an_error_t an_outchan_send(an_chan_t chan, an_packet_t packet);

/*
 * Allocate additional buffer memory for a channel.
 * The indicated number of pages of memory are allocated and broken
 * into buffers for receiving packets on an in-channel.  Along with
 * an_packet_return, this is how buffer memory is replenished.
 * Returns an error if applied to a dead channel or an out-channel.
 * XXX number of pages is probably not the right unit for this.
 *
 * ``chan'' is the channel to add memory to.
 * ``npages'' is the amount of memory to add.
 */
an_error_t an_inchan_addmem(an_chan_t chan, int npages);

#endif /* _NODEOS_CHANNEL_H_ */
