/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS Synchronization API.
 *
 * Mutexes and condition variables
 */
#ifndef _NODEOS_SYNCH_H_
#define _NODEOS_SYNCH_H_

#include <nodeos/impl.h>

typedef struct {
	int stuff;
} *an_mutex_t;

typedef struct {
	int stuff;
} *an_cond_t;

/* XXX compat */
#ifdef NODEOS_SPEC_COMPAT
#define an_Mutex an_mutex_t
#define an_mutexCreate an_mutex_create
#define an_mutexLock an_mutex_lock
#define an_mutexTrylock an_mutex_trylock
#define an_mutexUnlock an_mutex_unlock
#define an_mutexDestroy an_mutex_destroy
#define an_Cond an_cond_t
#define an_condCreate an_cond_create
#define an_condWait an_cond_wait
#define an_condSignal an_cond_signal
#define an_condBroadcast an_cond_broadcast
#define an_condDestroy an_cond_destroy
#endif

/*
 * Create a mutex object.
 * Mutex is unlocked upon return.  Returns a pointer or zero on error.
 */
an_mutex_t an_mutex_create(void);

/*
 * Lock a mutex object.
 * Blocks (uninterruptibly) if the mutex is held.
 */
void an_mutex_lock(an_mutex_t mutex);

/*
 * Attempt to lock a mutex.
 * Returns 0 if the lock suceeds, non-zero if the lock was already held.
 */
an_error_t an_mutex_trylock(an_mutex_t mutex);

/*
 * Unlocks a mutex.
 */
void an_mutex_unlock(an_mutex_t mutex);

/*
 * Destroys a mutex.
 * Fails and returns non-zero if the mutex is currently locked.
 */
an_error_t an_mutex_destroy(an_mutex_t mutex);

/*
 * Create a condition variable.
 * Returns a pointer or zero on error.
 */
an_cond_t an_cond_create(void);

/*
 * Wait for a condition to be signalled.
 * Atomically releases the specified mutex and blocks (interruptibly)
 * until signalled.  Spurious wakeups are possible and, in fact, will
 * occur if the thread is an_thread_intr'ed while in a wait.
 */
void an_cond_wait(an_cond_t cond, an_mutex_t mutex);

/*
 * Wakeup a single thread blocked on a condition variable.
 */
void an_cond_signal(an_cond_t cond);

/*
 * Wakeup all threads blocked on a condition variable.
 */
void an_cond_broadcast(an_cond_t cond);

/*
 * Destroy a condition variable.
 * Returns zero.
 */
an_error_t an_cond_destroy(an_cond_t cond);

#endif /* _NODEOS_SYNCH_H_ */
