/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Channel API
 */
#include <errno.h>
#include <nodeos/channel.h>

#include "nodeos.h"
#include "channel.h"
#include "demux_tlist.h"
#include "packet.h"
#include "mempool.h"

/*
 * Memory associated with a channel.
 * This includes not only the channel object itself, but any items referenced
 * by the channel object; i.e., non-scalar data that is passed across the API
 * by-reference.  Arguably, the latter should be part of the channel object
 * and we should just copyin directly to the appropriate fields here, or the
 * channel implementation could do the copyin calls, but we are trying to
 * preserve the tenuous separation between this ``entry layer'' and the
 * implementation.
 */
struct chandata {
	channel_t chan;		/* must be first */
	channel_proto_t proto;
	channel_addr_t addr;
};

/*
 * @deallocations: channel_t, channel_proto_t and channel_addr_t
 * returned to the chan's flow's mempool.
 */
static void
chan_cleanup(channel_t *chan)
{
	mempool_free(channel_mempool(chan), chan, sizeof(struct chandata));
}

/*
 * XXX still need to move channel buffer allocation up to here.
 *
 * @allocations: channel_t, channel_proto_t and channel_addr_t
 * charged to indicated flow.
 */
an_chan_t
an_inchan_create(an_flow_t flow, an_demuxkey_t key, an_chanproto_t *proto,
		 an_chanaddr_t *addr, an_memspec_t *mem, an_cpuspec_t *cpu,
		 an_chanrecvfunc_t func, void *farg)
{
	struct chandata *chandata = 0;
	flow_t *iflow;
	demux_tlist_t *ikey;
	cpuspec_t icpu;
	memspec_t imem;
	int err;

	NODEOS_ENTER;

	/*
	 * Validate arguments.
	 */
	iflow = NODEOS_INTERN_FLOW(flow);
	ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (iflow == 0 || proto == 0 || addr == 0 || mem == 0 || cpu == 0) {
		err = EINVAL;
		goto done;
	}

	/*
	 * Allocate memory.
	 */
	chandata = mempool_alloc(iflow->mpool, sizeof *chandata);
	if (chandata == 0) {
		err = ENOMEM;
		goto done;
	}
	memset((void *)chandata, 0, sizeof *chandata);

	/*
	 * Copy in data.
	 */
	NODEOS_COPYIN(&chandata->proto, proto, sizeof *proto, &err);
	if (err)
		goto done;
	NODEOS_COPYIN(&chandata->addr, addr, sizeof *addr, &err);
	if (err)
		goto done;
	NODEOS_COPYIN(&icpu, cpu, sizeof icpu, &err);
	if (err)
		goto done;
	NODEOS_COPYIN(&imem, mem, sizeof imem, &err);
	if (err)
		goto done;

	/*
	 * Create the channel.
	 */
	err = channel_create(&chandata->chan, CHAN_IN, iflow,
			     &chandata->proto, &chandata->addr,
			     ikey, icpu.max, imem.npages*NODEOS_PAGESIZE, 0,
			     func, farg, chan_cleanup);
 done:
	if (err) {
		if (chandata)
			mempool_free(iflow->mpool, chandata, sizeof *chandata);
	}
	NODEOS_LEAVE;
	return err ? 0 : NODEOS_EXTERN_CHAN(&chandata->chan);
}

/*
 * XXX net spec ignored for now.
 *
 * @allocations: channel_t, channel_proto_t and channel_addr_t
 * charged to indicated flow.
 */
an_chan_t
an_outchan_create(an_flow_t flow, an_chanproto_t *proto, an_chanaddr_t *addr,
		  an_netspec_t *net)
{
	struct chandata *chandata = 0;
	flow_t *iflow;
	int err;

	NODEOS_ENTER;

	/*
	 * Validate arguments.
	 */
	iflow = NODEOS_INTERN_FLOW(flow);
	if (iflow == 0 || proto == 0 || addr == 0) {
		err = EINVAL;
		goto done;
	}

	/*
	 * Allocate memory.
	 */
	chandata = mempool_alloc(iflow->mpool, sizeof *chandata);
	if (chandata == 0) {
		err = ENOMEM;
		goto done;
	}
	memset((void *) chandata, 0, sizeof *chandata);

	/*
	 * Copy in data.
	 */
	NODEOS_COPYIN(&chandata->proto, proto, sizeof *proto, &err);
	if (err)
		goto done;
	NODEOS_COPYIN(&chandata->addr, addr, sizeof *addr, &err);
	if (err)
		goto done;

	/*
	 * Create the channel.
	 */
	err = channel_create(&chandata->chan, CHAN_OUT, iflow,
			     &chandata->proto, &chandata->addr,
			     0, 0, 0, 0, 0, 0, chan_cleanup);
 done:
	if (err) {
		if (chandata)
			mempool_free(iflow->mpool, chandata, sizeof *chandata);
	}
	NODEOS_LEAVE;
	return err ? 0 : NODEOS_EXTERN_CHAN(&chandata->chan);
}

an_chan_t
an_cutchan_create(an_flow_t flow, an_chanproto_t *proto,
		  an_chan_t inchan, an_chan_t outchan)
{
	struct chandata *chandata = 0;
	flow_t *iflow;
	channel_t *iinchan, *ioutchan;
	int err;

	NODEOS_ENTER;

	/*
	 * Validate arguments.
	 */
	iflow = NODEOS_INTERN_FLOW(flow);
	iinchan = NODEOS_INTERN_CHAN(inchan);
	ioutchan = NODEOS_INTERN_CHAN(outchan);
	if (iflow == 0 || iinchan == 0 || ioutchan == 0) {
		err = EINVAL;
		goto done;
	}

	/*
	 * Allocate memory.
	 */
	chandata = mempool_alloc(iflow->mpool, sizeof *chandata);
	if (chandata == 0) {
		err = ENOMEM;
		goto done;
	}

	/*
	 * Copy in data.
	 */
	if (proto) {
		NODEOS_COPYIN(&chandata->proto, proto, sizeof *proto, &err);
		if (err)
			goto done;
	}

	/*
	 * Create the channel.
	 */
	err = cut_channel_create(&chandata->chan, iflow,
				 proto ? &chandata->proto : 0,
				 iinchan, ioutchan, cut_channel_receive,
				 chan_cleanup);
 done:
	if (err) {
		if (chandata)
			mempool_free(iflow->mpool, chandata, sizeof *chandata);
	}
	NODEOS_LEAVE;
	return err ? 0 : NODEOS_EXTERN_CHAN(&chandata->chan);

}

void
an_chan_destroy(an_chan_t chan)
{
	NODEOS_ENTER;
	channel_destroy(NODEOS_INTERN_CHAN(chan));
	NODEOS_LEAVE;
}

an_error_t
an_outchan_send(an_chan_t chan, an_packet_t packet)
{
	int err;

	NODEOS_ENTER;
	err = channel_send(NODEOS_INTERN_CHAN(chan),
			   NODEOS_INTERN_PACKET(packet));
	NODEOS_LEAVE;
	return (an_error_t)err;
}

/*
 * XXX need to move buffer allocation up to here.
 */
an_error_t
an_inchan_addmem(an_chan_t chan, int npages)
{
	int err;

	NODEOS_ENTER;
	err = channel_createbufs((channel_t *)chan, npages*NODEOS_PAGESIZE);
	NODEOS_LEAVE;
	return (an_error_t)err;
}

