/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Thread API
 */
#include <errno.h>
#include <nodeos/thread.h>

#include "nodeos.h"
#include "flow.h"
#include "threadpool.h"
#include "thread.h"

an_thread_t
an_thread_current(void)
{
	an_thread_t thread;

	NODEOS_ENTER;
	thread = (an_thread_t)thread_current();
	NODEOS_LEAVE;
	return thread;
}

an_error_t
an_thread_start(an_thread_t *out_thread, an_flow_t flow,
	       void (*func)(void *arg), void *arg)
{
	flow_t *iflow;
	thread_t *ith;
	int err;

	NODEOS_ENTER;

	if (flow == 0 || out_thread == 0 || func == 0) {
		err = EINVAL;
		goto done;
	}

	iflow = NODEOS_INTERN_FLOW(flow);
	if (iflow != 0) {
		err = threadpool_thread_start(iflow->tpool, 0,
					      (thread_func_t)func, arg, 0,
					      &ith);
		if (err == 0)
			*out_thread = NODEOS_EXTERN_THREAD(ith);
	}

 done:
	NODEOS_LEAVE;
	return err;
}

void
an_thread_exit(void)
{
	NODEOS_ENTER;
	thread_exit();
	NODEOS_LEAVE;
}

void
an_thread_yield(void)
{
	NODEOS_ENTER;
	thread_yield();
	NODEOS_LEAVE;
}

void
an_thread_sleep(an_timespec_t *delay)
{
	NODEOS_ENTER;
	thread_sleep(delay);
	NODEOS_LEAVE;
}

void
an_thread_setdata(void *data)
{
	NODEOS_ENTER;
	thread_setdata(data);
	NODEOS_LEAVE;
}

void *
an_thread_getdata(void)
{
	void *data;

	NODEOS_ENTER;
	data = thread_getdata();
	NODEOS_LEAVE;

	return data;
}

void
an_thread_intr(an_thread_t thread)
{
	thread_t *ithread;

	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread != 0)
		thread_interrupt(ithread);
	NODEOS_LEAVE;
}

void
an_thread_blockintr(void)
{
	NODEOS_ENTER;
	thread_block_interrupt();
	NODEOS_LEAVE;
}

void
an_thread_unblockintr(void)
{
	NODEOS_ENTER;
	thread_unblock_interrupt();
	NODEOS_LEAVE;
}

an_error_t
an_thread_setprio(an_thread_t thread, int prio)
{
	thread_t *ithread;
	int err;

	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread == 0)
		err = EINVAL;
	else
		err = thread_setprio(ithread, prio);
	NODEOS_LEAVE;

	return (an_error_t) err;
}

an_error_t
an_thread_getprio(an_thread_t thread, int *priop)
{
	thread_t *ithread;
	int iprio, err;

	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread == 0)
		err = EINVAL;
	else {
		iprio = thread_getprio(ithread);
		if (err == 0)
			NODEOS_COPYOUT(priop, &iprio, sizeof *priop, &err);
	}
	NODEOS_LEAVE;

	return (an_error_t) err;
}

/*
 * Implementation specific API calls
 */

void
an_thread_sethandler(an_thread_intr_t handler)
{
	NODEOS_ENTER;
	thread_sethandler(handler);
	NODEOS_LEAVE;
}
