/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A raw ethernet channel driver.
 */
#include "options.h"

#ifdef CHAN_ETH

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#ifdef OSKIT
#include <oskit/startup.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/ethernet.h>
#include <oskit/dev/net.h>
#include <oskit/net/ether.h>
#include <oskit/io/bufio.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#ifndef ETHER_HDR_SIZE
#define ETHER_HDR_SIZE 14
#endif

#include "unix_eth.h"
#endif

#include "nodeos.h"
#include "os.h"
#include "channel.h"
#include "demux.h"
#include "demux_tlist.h"
#include "packet.h"
#include "thread.h"
#include "ilock.h"
#include "mem.h"
#include "misc.h"

struct chandata;

/*
 * Per-device structure.
 * Most oskit drivers restrict to only a single open of a device,
 * we use a single send/recv netio shared by many channels.
 *
 * The etherdev lock is lower in the locking hierarchy than the channel
 * lock.
 */
struct etherdev {
	ilock_t		 lock;
	char		 name[NODEOS_MAXIFNAME]; /* "eth0", etc. */
	unsigned char	 addr[6];	/* our address */
	demux_t		 *demux;	/* list of demuxers for this device */
	char		 recv_inuse;	/* Flag */
	char		 send_inuse;	/* Flag */

#ifdef OSKIT        
	/* OSKit specific */
	oskit_etherdev_t *eth;		/* device handle */
	oskit_netio_t	 *send;		/* handle for outgoing packets */
	oskit_netio_t	 *recv;		/* handle for incoming packets */
#else
	char		if_name[IFNAMSIZ]; /* system ethernet device name */
	int		recv;           /* socket */
#endif        
};

/*
 * Ethernet channel private data.  Protected by the channel lock.
 * Contains a pointer to the etherdev structure and, for inChan channels,
 * an array of received packet descriptors.
 *
 * The packet descriptors are a hack for passing information between the
 * interrupt handler and the channel thread which processes the packet.
 * Two alternatives: 1) allow passing one more argument to thread startup
 * so that we could pass chan/data/size directly, 2) wrap the data in a
 * NodeOS packet_t at interrupt time and hand that as the second argument.
 * The latter is the better approach and what we will ultimately do, but
 * we need more infrastructure (at least for the OSKit) to enable us to
 * perform a buffer swap so that the channel "pays" for the packet
 * immediately.
 */
struct chandata {
	channel_t	 *chan;		/* channel we are a part of */
	struct etherdev	 *edev;		/* associated ether device */
	struct chanpdata {
		void		*data;
		size_t		size;
	}		 *free;		/* packet descriptor free list */
	struct chanpdata pdata[0];	/* zero or more packet descriptors */
};

/* XXX this should possibly be exposed */
struct chanethaddr {
	unsigned char addr[6];
	unsigned short type;
};


static struct etherdev *etherdevs;
static int netherdevs;
static char chanprefix[] = "eth";
static ilock_t ethlock;

static struct etherdev *chan_edev_find(char *name);
static int chan_edev_demux_add(struct etherdev *edev, channel_t *chan);
static void chan_edev_demux_remove(struct etherdev *edev, channel_t *chan);
static int chan_edev_init(struct etherdev *edev);
static int chan_edev_deinit(struct etherdev *edev);
static void chan_eth_receive(void *arg1, void *arg2);

#ifdef OSKIT
static oskit_error_t chan_edev_intr(void *arg, oskit_bufio_t *b,
				    oskit_size_t psize);

extern oskit_netio_t *oskit_threaded_netio_create(
  oskit_error_t (*func)(void *data, oskit_bufio_t *b, oskit_size_t pkt_size),
  void *data);
#endif

void
chan_eth_init(void)
{
	int ndev, i;
#ifdef OSKIT
	oskit_etherdev_t **edev;
#else
	struct if_info *edev;
#endif        

	ilock_init(&ethlock);
	ilock_lock(&ethlock);

	/*
	 * Look up all the ethernet devices
	 */
#ifdef OSKIT
	ndev = osenv_device_lookup(&oskit_etherdev_iid, (void***)&edev);
#else 
	ndev = unix_eth_get_if_info(&edev);
#endif
	if (ndev <= 0) {
                printf("%s: no devices!\n", __FUNCTION__);
		return;
        }

	/*
	 * Create a device to name mapping
	 */
	etherdevs = mem_alloc(ndev * sizeof(struct etherdev));
	if (etherdevs == 0)
		return;
	memset(etherdevs, 0, ndev * sizeof(struct etherdev));
	for (i = 0; i < ndev; i++) {
		snprintf(etherdevs[i].name, sizeof etherdevs[i].name,
			 "%s%d", chanprefix, i);
#ifdef OSKIT
		etherdevs[i].eth = edev[i];
		oskit_etherdev_getaddr(edev[i], etherdevs[i].addr);
#else 
		strncpy(etherdevs[i].if_name, edev[i].name, IFNAMSIZ);
		memcpy(etherdevs[i].addr, edev[i].addr, 6);
#endif

		ilock_init(&etherdevs[i].lock);
	}
	netherdevs = ndev;
	free(edev);

	ilock_unlock(&ethlock);
}

int
chan_eth_info(an_ifinfo_t info[], int *count)
{
	int i, num;

	assert(count != 0);

	num = *count;
	if (num == 0) {
		*count = netherdevs;
		return 0;
	}
	if (num > netherdevs)
		num = netherdevs;

	/*
	 * Fill in information about the interface.
	 * Note that the info id is assigned elsewhere and that
	 * the info ethaddr saddr and type are zero (indicating "any").
	 */
	memset(info, 0, sizeof info[0] * num);
	ilock_lock(&ethlock);
	for (i = 0; i < num; i++) {
		info[i].type = ANIF_ETH;
		strncpy(info[i].pname, etherdevs[i].name,
			sizeof etherdevs[i].name);
		memcpy(info[i].addr.eth.daddr, etherdevs[i].addr,
		       sizeof etherdevs[i].addr);
	}
	ilock_unlock(&ethlock);

	*count = num;
	return 0;
}

/*
 * Hook up to the ethernet device.
 * Called with the channel locked.
 *
 * @allocation: ethernet specific channel data structure
 */
int
chan_eth_create(channel_t *chan)
{
	struct chandata *cdata;
	struct etherdev *edev;
	size_t size;
	int err;

	/* XXX not yet */
	if (channel_type(chan) == CHAN_CUT)
		return EINVAL;

	/*
	 * Lookup the device corresponding to the name given
	 */
	edev = chan_edev_find(channel_name(chan));
	if (edev == 0)
		return EINVAL;

	/*
	 * Create the ethernet specific data struct.
	 * For in-chans, this includes one packet descriptor per possible
	 * service thread.  This is a gross hack and potentially incorrectly
	 * limits the in-chan (i.e., if it has more packet buffers than
	 * service threads, it should be able to receive more packets than
	 * we allow here).
	 */
	size = sizeof(struct chandata);
	if (channel_type(chan) == CHAN_IN)
		size += sizeof(struct chanpdata) * channel_maxthreads(chan);
	cdata = mempool_alloc(channel_mempool(chan), size);
	if (cdata == 0)
		return ENOMEM;

	memset(cdata, 0, size);
	cdata->edev = edev;

	/*
	 * Initialize the packet descriptor free list
	 */
	if (channel_type(chan) == CHAN_IN) {
		struct chanpdata *cdp;

		cdp = (struct chanpdata *)((void *)cdata + size);
		while (--cdp >= cdata->pdata) {
			cdp->data = cdata->free;
			cdata->free = cdp;
		}
	}

	/*
	 * Now hook up to the ethernet device
	 */
	ilock_lock(&edev->lock);

	/*
	 * If device hasn't been initialized, do so now
	 */
	if (edev->recv == 0) {
		err = chan_edev_init(edev);
		if (err) {
			ilock_unlock(&edev->lock);
			mempool_free(channel_mempool(chan), cdata, size);
			return err;
		}
	}
	assert(edev->recv);

	/*
	 * Hook up channel and local data
	 */
	chan->data = cdata;
	cdata->chan = chan;

	/*
	 * Add the demux filter based on the address and demux info.
	 */
	if (channel_type(chan) == CHAN_IN) {
		err = chan_edev_demux_add(edev, chan);
		if (err) {
			ilock_unlock(&edev->lock);
			mempool_free(channel_mempool(chan), cdata, size);
			return err;
		}
	}

	/*
	 * For deallocation. 
	 */
	if (channel_type(chan) == CHAN_IN)
		edev->recv_inuse++;
	else
		edev->send_inuse++;

	ilock_unlock(&edev->lock);
	return 0;
}

/*
 * Called with the channel locked
 */
void
chan_eth_destroy(channel_t *chan)
{
	struct chandata *cdata;
	struct etherdev *edev;
	size_t size;

	cdata = chan->data;
	size = sizeof(struct chandata);
	if (channel_type(chan) == CHAN_IN)
		size += sizeof(struct chanpdata) * channel_maxthreads(chan);

	edev = cdata->edev;
	ilock_lock(&edev->lock);

	/*
	 * Destroy underlying goo when there are no more channels using it.
	 */
	if (channel_type(chan) == CHAN_IN)
		edev->recv_inuse--;
	else
		edev->send_inuse--;

	assert(edev->send_inuse >= 0);
	assert(edev->recv_inuse >= 0);

	if (edev->recv_inuse == 0 && edev->send_inuse == 0)
		chan_edev_deinit(edev);

	/* XXX kill service threads, free up packets/bufios */
	
	/*
	 * Remove the demux
	 */
	if (channel_type(chan) == CHAN_IN && chan->tdata.in.demux)
		chan_edev_demux_remove(edev, chan);

	ilock_unlock(&edev->lock);

	mempool_free(channel_mempool(chan), cdata, size);
}

/*
 * Send a packet out on a channel.
 * Called with the channel locked.
 */
void
chan_eth_send(channel_t *chan, packet_t *pack)
{
#ifdef OSKIT
	oskit_bufio_t *bio;
	oskit_size_t off;
	int err;
	void *data;        
	size_t size;
#endif
	struct ether_header eth, *ethp; 
	struct etherdev *edev;
	an_ethaddr_t *eaddr;
        
	edev = ((struct chandata *)(chan->data))->edev;

#ifdef OSKIT        
	/*
	 * Create a bufio for the packet and header.
	 */
	size = packet_size(pack);
	if ((chan->flags & CHAN_ISRAW) == 0)
		size += ETHER_HDR_SIZE;
	bio = oskit_bufio_create(size);
	if (bio == 0)
		return;

	if ((chan->flags & CHAN_ISRAW) == 0) {
		eaddr = &chan->addr->eth;
	} else
		eaddr = 0;

	/*
	 * Copy header info and data into the bufio.
	 * We do this by mapping if possible, by writing otherwise.
	 */
	err = oskit_bufio_map(bio, &data, 0, size);
	if (err == 0) {
		if (eaddr) {
			ethp = data;
			memcpy(ethp->ether_dhost, eaddr->daddr,
			       sizeof ethp->ether_dhost);
			memcpy(ethp->ether_shost, edev->addr,
			       sizeof ethp->ether_shost);
			ethp->ether_type = eaddr->type;
			memcpy(data + ETHER_HDR_SIZE, packet_data(pack),
			       packet_size(pack));
		} else
			memcpy(data, packet_data(pack), packet_size(pack));
		oskit_bufio_unmap(bio, data, 0, size);
	} else {
		if (eaddr) {
			memcpy(eth.ether_dhost, eaddr->daddr,
			       sizeof eth.ether_dhost);
			memcpy(eth.ether_shost, edev->addr,
			       sizeof eth.ether_shost);
			eth.ether_type = eaddr->type;
			err = oskit_bufio_write(bio, &eth, 0, ETHER_HDR_SIZE,
						&off);
			if (err != 0 || off != ETHER_HDR_SIZE) {
				oskit_bufio_release(bio);
				return;
			}
		} else
			off = 0;
		err = oskit_bufio_write(bio, packet_data(pack), off,
					packet_size(pack), &off);
		if (err != 0 || off != packet_size(pack)) {
			oskit_bufio_release(bio);
			return;
		}
	}

	/*
	 * Push the packet out.
	 * We lock the edev here but that is probably unnecessary as we
	 * are serialized by the channel lock.
	 */
	ilock_lock(&edev->lock);
	(void) oskit_netio_push(edev->send, bio, size);
	ilock_unlock(&edev->lock);
	oskit_bufio_release(bio);
#else
	if ((chan->flags & CHAN_ISRAW) == 0) {
		eaddr = &chan->addr->eth;

		memcpy(&eth.ether_dhost, eaddr->daddr, sizeof eth.ether_dhost);
		memcpy(&eth.ether_shost, edev->addr, sizeof eth.ether_shost);
		eth.ether_type = eaddr->type;
		ethp = &eth;
	} else
		ethp = 0;
	os_write_packet(edev->recv, packet_data(pack), packet_size(pack), ethp);
#endif /* OSKIT */        
}

static struct etherdev *
chan_edev_find(char *name)
{
	int i, len;

	len = strlen(name);

	ilock_lock(&ethlock);
	for (i = 0; i < netherdevs; i++)
		if (strncmp(etherdevs[i].name, name, len) == 0) {
			ilock_unlock(&ethlock);
			return &etherdevs[i];
		}

	ilock_unlock(&ethlock);
	return 0;
}

/*
 * Register channel demux information with the device.  We incorporate the
 * address and type information into the demux as well.
 * Called with the channel and etherdev locked.
 */
static int
chan_edev_demux_add(struct etherdev *edev, channel_t *chan)
{
	an_ethaddr_t *eaddr;
	demux_tlist_t *list = 0;
	int err, n, flag;
	static unsigned char zaddr[6] = { 0 };

	assert(channel_type(chan) == CHAN_IN);

	/*
	 * If RAW, just register the provided demux
	 */
	if (chan->flags & CHAN_ISRAW)
		goto done;

	eaddr = &chan->addr->eth;

	/*
	 * Determine how many demux elements we are adding.
	 * XXX is it safe to use 0 addresses as wildcards?
	 */
	n = flag = 0;
	if (memcmp(eaddr->daddr, &zaddr, sizeof zaddr) != 0)
		n++, flag |= 1;
	if (memcmp(eaddr->saddr, &zaddr, sizeof zaddr) != 0)
		n++, flag |= 2;
	if (eaddr->type != 0)
		n++, flag |= 4;
	if (n == 0 && chan->tdata.in.demux == 0)
		goto done;

	/*
	 * Expand the existing demux list (or create a new one)
	 */
	if (chan->tdata.in.demux == 0)
		err = demux_tlist_create(n, &list);
	else {
		err = demux_tlist_copy(chan->tdata.in.demux,
				       chan->tdata.in.demux->ntriples+n, &list);
		if (err == 0)
			demux_tlist_adjust(list, ETHER_HDR_SIZE);
	}
	if (err)
		return err;

	/*
	 * Tack on matches for the address and type information
	 */
	if ((flag & 1) == 0)
		err = demux_tlist_prepend(list, 0, 6, eaddr->daddr);
	if ((flag & 2) != 0 && err == 0)
		err = demux_tlist_prepend(list, 6, 6, eaddr->saddr);
	if ((flag & 4) != 0 && err == 0)
		err = demux_tlist_prepend(list, 12, 2, &eaddr->type);
	if (err) {
		demux_tlist_destroy(list);
		return err;
	}

	/*
	 * Replace the old channel demux info with our concoction
	 * and register the new info.
	 *
	 * XXX deref the original demuxkey object?
	 */
	chan->tdata.in.demux = list;
 done:
	err = channel_demux_add(chan, &edev->demux);
	if (err) {
		if (list)
			demux_tlist_destroy(list);
		chan->tdata.in.demux = 0;
	}

	return err;
}

static void
chan_edev_demux_remove(struct etherdev *edev, channel_t *chan)
{
	assert(channel_type(chan) == CHAN_IN);

	channel_demux_remove(chan, &edev->demux);
	demux_tlist_destroy(chan->tdata.in.demux);
	chan->tdata.in.demux = 0;
}

#ifdef OSKIT
/*
 * Receive netio push function, called from the interrupt handler but in
 * a full thread context.  Bufio comes in with a ref we must release to
 * free the buffer.
 *
 * This routine serializes all packets coming in on the netio (currently one
 * netio per interface) so lets get in and out as quick as possible.  Just
 * queue the bufio and fire up a flow thread.
 */
static oskit_error_t
chan_edev_intr(void *arg, oskit_bufio_t *b, oskit_size_t psize)
{
	struct etherdev *edev = arg;
	channel_t *chan;
	struct chandata *cdata;
	struct chanpdata *cdp;
	void *data;

	/*
	 * Map the packet for demux examination
	 */
	assert(psize >= ETHER_HDR_SIZE);
	if (oskit_bufio_map(b, &data, 0, psize) != 0) {
		oskit_bufio_release(b);
		return 0;
	}

	/*
	 * Find a channel for this packet.
	 * If no channel matches, the packet is dropped.
	 */
	ilock_lock(&edev->lock);
	chan = demux_apply(edev->demux, data, psize);
	ilock_unlock(&edev->lock);
	oskit_bufio_unmap(b, data, 0, psize);
	if (chan == 0) {
		oskit_bufio_release(b);
		return 0;
	}

	channel_lock(chan);

	/*
	 * XXX FIXME FIXME
	 * Channel has no receive function, it is going to be associated
	 * with a cutchan but hasn't yet.  We shouldn't catch this here.
	 * Instead, we should not register a demux for the channel until
	 * the cutchan is created (and the function set).
	 */
	if (chan->tdata.in.func == 0) {
		channel_unlock(chan);
		oskit_bufio_release(b);
		return 0;
	}

	/* XXX channel shutdown has begun */
	if (chan->flags & CHAN_DYING) {
		channel_unlock(chan);
		oskit_bufio_release(b);
		return 0;
	}

	/*
	 * If there are no packet descriptors available, or we cannot
	 * start a service thread, we just drop the packet.
	 */
	cdata = chan->data;
	cdp = cdata->free;
	if (cdp == 0 || chan->curthreads >= channel_maxthreads(chan) ||
	    threadpool_thread_schedule(chan->tpool, chan, 1,
				       chan_eth_receive, chan, cdp) != 0) {
		channel_unlock(chan);
		oskit_bufio_release(b);
		return EAGAIN; /* XXX */
	}
	chan->curthreads++;
	cdata->free = cdp->data;

	/*
	 * Queue the bufio/size and get out.  Releasing the lock will allow
	 * the flow thread to do its work.
	 */
	cdp->data = b;
	cdp->size = psize;
	channel_unlock(chan);

	return 0;
}

#else

void *
chan_edev_read_loop(void *arg)
{
        struct etherdev *edev = arg;
        int count;
        char *buf;
        int bufsize;
	channel_t *chan;
	struct chandata *cdata;
	struct chanpdata *cdp;

	bufsize = NODEOS_CHAN_BUFSIZE;

	while (1) {
		buf = mem_alloc(bufsize);
		assert(buf != 0);

	again:
		count = os_read_packet(edev->recv, buf, bufsize);
		if (count < 0) {
			printf("%s: reader thread died: %s\n",
			       edev->if_name, strerror(errno));
			panic("NodeOS: enet interface error");
		}

		/* XXX should we delay here? */
		if (count == 0)
			goto again;
		
		/*
		 * Find a channel for this packet.
		 * If no channel matches, the packet is dropped.
		 */
		ilock_lock(&edev->lock);
		chan = demux_apply(edev->demux, buf, count);
		ilock_unlock(&edev->lock);
		if (chan == 0)
			goto again;

		channel_lock(chan);

		/* XXX channel shutdown has begun */
		if (chan->flags & CHAN_DYING) {
			channel_unlock(chan);
			break;
		}

		/*
		 * No available buffer slot or a failure to start a thread
		 * indicates no CPU resources.  Just drop the packet.
		 */
		cdata = chan->data;
		cdp = cdata->free;
		if (cdp == 0 ||
		    threadpool_thread_schedule(chan->tpool, chan, 1,
					       chan_eth_receive, chan, cdp)) {
			channel_unlock(chan);
			goto again;
		}
		cdata->free = cdp->data;

		/*
		 * Otherwise, queue the buffer and get back to work.
		 * Releasing the lock will allow the flow thread to do
		 * its work.  The flow thread will free the buffer when
		 * it is done.
		 */
		cdp->data = buf;
		cdp->size = count;
                channel_unlock(chan);
	}

	mem_free(buf, bufsize);
        return 0;
}
#endif /* OSKIT */

static int
chan_edev_init(struct etherdev *edev)
{
	int err;
#ifndef OSKIT
        pthread_t id;
#endif        
        
#ifdef OSKIT        
	/*
	 * Create the netio object we will use for receiving packets.
	 */
	edev->recv = oskit_threaded_netio_create(chan_edev_intr, edev);
	if (edev->recv == 0)
		return ENOMEM; /* XXX */

	/*
	 * Open the device registering our receive netio and getting
	 * back the send netio.
	 */
	assert(edev->send == 0);
	err = oskit_etherdev_open(edev->eth, 0, edev->recv, &edev->send);
	if (err) {
		oskit_netio_release(edev->recv);
		return err;
	}
#else
	edev->recv = os_open_eth(edev->if_name);
	if (edev->recv < 0) 
		return EINVAL;
        
	err = pthread_create(&id, 0, chan_edev_read_loop, (void *)edev);
	if (err)
		return err;
	err = pthread_detach(id);
	assert(err == 0);
#endif        
	return 0;
}

/*
 * Gotta release the ethernet devices or my PC hangs up in netboot reboot.
 */
static int
chan_edev_deinit(struct etherdev *edev)
{
#ifdef OSKIT
	oskit_netio_release(edev->send);
	oskit_netio_release(edev->recv);
	oskit_etherdev_release(edev->eth);

	/* XXX This will cause another trip through chan_edev_init() */
	edev->send = edev->recv = 0;
#endif        
	return 0;
}

/*
 * This is where we startup in a flow thread which handles a packet.
 */
static void
chan_eth_receive(void *arg1, void *arg2)
{
	channel_t *chan;
	struct chandata *cdata;
	struct chanpdata *cdp;
#ifdef OSKIT
	oskit_bufio_t *b;
#else
	struct chanbuf *cb;
#endif
	size_t psize, hsize;
	void *data;
	flow_t *flow;
	packet_t *packet;
	channel_recv_t func;
	void *farg;
	int err, iscutchan;
        
	chan = arg1;
	assert_channel_valid(chan);

	channel_lock(chan);

	/* XXX channel shutdown has begun */
	if (chan->flags & CHAN_DYING) {
		channel_unlock(chan);
		return;
	}

	/*
	 * Extract the relevant info and release the packet descriptor.
	 */
	cdp = arg2;
	assert(cdp != 0);
#ifdef OSKIT
	b = cdp->data;
#else
	data = cdp->data;
#endif        
	psize = cdp->size;
	cdata = chan->data;
	cdp->data = cdata->free;
        cdata->free = cdp;

	flow = chan->flow;
	func = chan->tdata.in.func;
	farg = chan->tdata.in.farg;
	iscutchan = (chan->flags & CHAN_WITHCUT);
	channel_unlock(chan);

	/*
	 * Copy data into a packet.
	 * XXX should do a bufio/aggregate kinda thing.
	 */
	hsize = (chan->flags & CHAN_ISRAW) ? 0 : ETHER_HDR_SIZE;
#ifdef OSKIT
	psize -= hsize;
	err = oskit_bufio_map(b, &data, hsize, psize);
	if (err) {
		oskit_bufio_release(b);
		return;
	}
	err = channel_packet_create(chan, data, psize, 0, 1, &packet, 0);
	oskit_bufio_unmap(b, data, 0, psize);
	oskit_bufio_release(b);
#else
	/*
	 * Create a packet from the receive buffer.
	 * XXX packet_create hands us back a replacement buffer which
	 * we have no way of using at the moment, so we just free it.
	 *
	 * XXX note inconsistant use of mempool_* and mem_*, we need to
	 * transfer memory between pools at some point.
	 */
	err = channel_packet_create(chan, data, psize, hsize, 0,
				    &packet, &cb);
	mem_free(cb, NODEOS_CHAN_BUFSIZE);
#endif        
        if (err) 
		return;

	/*
	 * Finally!  We call the application supplied receive function.
	 */
	if (iscutchan) {
		/*
		 * XXX special case for cut-through channels,
		 * The function's first argument in this case is an internal
		 * packet, we just cast it to an an_packet_t to match the
		 * prototype for the receive function (me bad!).  The second
		 * argument, farg, is the internal cutchan object.
		 */
		func((an_packet_t)packet, farg);
	} else {
		NODEOS_LEAVE;
		func(NODEOS_EXTERN_PACKET(packet), farg);
		NODEOS_ENTER;
	}
}

#endif /* CHAN_ETH */
