/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A "socket" channel driver.
 *
 * Currently we recognize only "socket/udp/ip/eth".
 */
#include "options.h"

#ifdef CHAN_SOCKET

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef OSKIT
#include <oskit/startup.h>
#include <oskit/com/listener.h>
#include <oskit/dev/dev.h>
#include <oskit/net/socket.h>
#include <oskit/dev/timer.h>
#endif

#include "flow.h"
#include "channel.h"
#include "demux.h"
#include "packet.h"
#include "thread.h"
#include "synch.h"
#include "mem.h"

/*
 * One structure per socket.
 * Sockets may be shared by multiple channels, with each channel having
 * a different demux for inChans.
 */
struct socketdata {
	struct socketdata  *next;
	short		   stype; /* SOCK_DGRAM or SOCK_STREAM */
	short		   ctype; /* inChan or outChan */
	struct sockaddr	   addr;  /* IP addr and UDP/TCP port */
	int		   fd;	  /* socket file descriptor */
	demux_t		   *demux;/* list of demuxers for this socket */
	struct chanbuf	   *rbuf; /* receive buffer */
	int		   refs;  /* refcount */
	ilock_t		   lock;  /* protects the demux list and rbuf */
};

static struct socketdata *socketlist;
static ilock_t socketlock;
static icond_t socketcond;

/*
 * Channel specific data (chan->data) is a pointer to a socketdata
 * structure (inChan) or a socket file descriptor (outChan).
 */

static int chan_socketdata_find(int stype, an_sockaddr_t *caddr,
				int ctype, struct socketdata **out_sdata);
void chan_socketdata_release(struct socketdata *sdata);
static void chan_socketdata_intr(void *arg, void *arg2);

/*
 * In a couple of places we need to create "uncharged" resources.
 * These are accounted against the root flow.
 */
extern flow_t *root_flow;

void
chan_socket_init(void)
{
	ilock_init(&socketlock);
	icond_init(&socketcond);

#ifdef OSKIT        
	ilock_lock(&socketlock);
	start_network_router_pthreads();
	ilock_unlock(&socketlock);
#endif
}

int
chan_socket_create(channel_t *chan)
{
	struct socketdata *sdata;
	an_sockaddr_t *caddr;
	int err, stype, len;
	char *cp;

	/* XXX */
	if (channel_type(chan) == CHAN_CUT || chan->addr == 0)
		return EINVAL;

	cp = channel_name(chan);
	len = channel_namelen(chan);
	if (strncmp(cp, "socket/udp/ip/eth", len) == 0)
		stype = SOCK_DGRAM;
#if 0
	else if (strncmp(cp, "socket/tcp/ip/eth", len) == 0)
		stype = SOCK_STREAM;
#endif
	else
		return EINVAL;

	caddr = &chan->addr->socket;

	/*
	 * Find/create a socket of the appropriate type.
	 * Returns the socket data locked.
	 */
	err = chan_socketdata_find(stype, caddr, channel_type(chan), &sdata);
	if (err)
		return err;

	/*
	 * Add the demux filter
	 */
	if (channel_type(chan) != CHAN_OUT) {
		err = channel_demux_add(chan, &sdata->demux);
		if (err) {
			chan_socketdata_release(sdata);
			return err;
		}
	}

	ilock_unlock(&sdata->lock);

	chan->data = sdata;
	return 0;
}

void
chan_socket_destroy(channel_t *chan)
{
	struct socketdata *sdata;

	sdata = chan->data;

	ilock_lock(&sdata->lock);
	channel_demux_remove(chan, &sdata->demux);
	chan_socketdata_release(sdata);
}

void
chan_socket_send(channel_t *chan, packet_t *pack)
{
	struct socketdata *sdata;

	sdata = chan->data;
	ilock_lock(&sdata->lock);
	(void) write(sdata->fd, packet_data(pack), packet_size(pack));
	ilock_unlock(&sdata->lock);
}

/*
 * Lookup or create a socket corresponding to the desired type/address.
 * Returns 0 on success with the socketdata struct locked.
 */
static int
chan_socketdata_find(int stype, an_sockaddr_t *caddr, int ctype,
		     struct socketdata **out_sdata)
{
	struct socketdata *sdata;
	int err;

	/*
	 * Look for an existing socket
	 */
	ilock_lock(&socketlock);
	for (sdata = socketlist; sdata != 0; sdata = sdata->next)
		if (sdata->ctype == ctype && sdata->stype == stype &&
		    memcmp(&sdata->addr, caddr, sizeof sdata->addr) == 0) {
			ilock_lock(&sdata->lock);
			ilock_unlock(&socketlock);
			sdata->refs++;
			*out_sdata = sdata;
			return 0;
		}
	ilock_unlock(&socketlock);

	/*
	 * Otherwise, create a new one
	 */
	sdata = mem_alloc(sizeof *sdata);
	if (sdata == 0)
		return ENOMEM;

	sdata->fd = socket(PF_INET, stype, 0);
	if (sdata->fd < 0) {
		mem_free(sdata, sizeof *sdata);
		return errno;
	}

	/*
	 * Allocate a receive buffer
	 */
	sdata->rbuf = mem_alloc(sizeof *sdata->rbuf);
	if (sdata->rbuf == 0) {
		close(sdata->fd);
		mem_free(sdata, sizeof *sdata);
		return errno;
	}

	/*
	 * We initialize and lock the sdata lock so we can use it to
	 * synchronize with any reader thread.  We set the ref count now
	 * so that we can use socketdata_release to cleanup on errors.
	 */
	ilock_init(&sdata->lock);
	ilock_lock(&sdata->lock);

	if (ctype == CHAN_IN) {
		if (bind(sdata->fd, caddr, sizeof *caddr) < 0) {
			err = errno;
			goto error;
		}

		err = threadpool_thread_start(root_flow->tpool, 1,
					      chan_socketdata_intr, sdata, 0,
					      0);
		if (err)
			goto error;
	} else {
		if (connect(sdata->fd, caddr, sizeof *caddr) < 0) {
			err = errno;
			goto error;
		}
	}

	/*
	 * Initialize the rest of the socket data
	 *
	 * Note that we acquire the socketlist lock out of order here.
	 * This is safe since no one else could hold our lock yet and cause
	 * deadlock.
	 */
	sdata->stype = stype;
	sdata->ctype = ctype;
	sdata->addr = *caddr;
	sdata->demux = 0;
	sdata->refs = 1;
	ilock_lock(&socketlock);
	sdata->next = socketlist;
	socketlist = sdata;
	ilock_unlock(&socketlock);

	*out_sdata = sdata;
	return 0;

error:
	close(sdata->fd);
	mem_free(sdata->rbuf, sizeof *sdata->rbuf);
	ilock_unlock(&sdata->lock);
	ilock_deinit(&sdata->lock);
	mem_free(sdata, sizeof *sdata);
	return err;
}

/*
 * Drop a reference to a socketdata struct.
 * Called with the sdata locked.  If no more refs, the sdata is destroyed.
 * Otherwise, the lock is dropped.
 */
void
chan_socketdata_release(struct socketdata *sdata)
{
	struct socketdata **sdp;

	if (--sdata->refs == 0) {
		/*
		 * If we cannot grab the list lock, do the funky lock dance.
		 * Drop our lock to get the global lock then reaquire ours.
		 * Then check to make sure someone didn't latch on to our
		 * sdata in the meantime.
		 */
		if (ilock_trylock(&socketlock)) {
			ilock_unlock(&sdata->lock);
			ilock_lock(&socketlock);
			ilock_lock(&sdata->lock);
			if (sdata->refs != 0) {
				ilock_unlock(&sdata->lock);
				ilock_unlock(&socketlock);
				return;
			}
		}

		/*
		 * Locate and remove from the list
		 */
		for (sdp = &socketlist; *sdp != 0; sdp = &(*sdp)->next)
			if (*sdp == sdata)
				break;
		assert(*sdp != 0);
		*sdp = sdata->next;
		ilock_unlock(&socketlock);

		/*
		 * Shutdown and destroy the socket
		 */
		close(sdata->fd);
		mem_free(sdata->rbuf, sizeof *sdata->rbuf);
		ilock_unlock(&sdata->lock);
		ilock_deinit(&sdata->lock);
		mem_free(sdata, sizeof *sdata);
	} else
		ilock_unlock(&sdata->lock);
}

/*
 * Receiver thread for inChan sockets.
 * Applies the demux filters for the socket to determine a channel to
 * deliver to and fires up a channel thread to deal with it.
 *
 * XXX for TCP support, do listen and accept here.
 */
static void
chan_socketdata_intr(void *arg, void *arg2)
{
	struct socketdata *sdata = arg;
	ssize_t size;
	int s, err;
	packet_t *packet;
	channel_t *chan;

	/*
	 * Lock forces synchronization with our creator
	 */
	ilock_lock(&sdata->lock);
	s = sdata->fd;
	ilock_unlock(&sdata->lock);

	/*
	 * Loop receiving and processing data
	 */
	while (1) {
		size = recvfrom(s, sdata->rbuf, sizeof *sdata->rbuf, 0, 0, 0);
		if (size < 0)
			break;

		/*
		 * Locate the target channel for this packet
		 */
		ilock_lock(&sdata->lock);
		chan = demux_apply(sdata->demux, sdata->rbuf, size);

		/*
		 * XXX FIXME FIXME
		 * Channel has no receive function, it is going to be associated
		 * with a cutchan but hasn't yet.  We shouldn't catch this here.
		 * Instead, we should not register a demux for the channel until
		 * the cutchan is created (and the function set).
		 */
		if (chan && chan->tdata.in.func == 0) {
			ilock_unlock(&sdata->lock);
			continue;
		}

		if (chan != 0)
			err = channel_packet_create(chan, sdata->rbuf, size, 0,
						    0, &packet, &sdata->rbuf);
		else
			err = 0;
		ilock_unlock(&sdata->lock);

		/*
		 * And fire up the handler
		 */
		if (chan && err == 0)
			channel_thread_start(chan, packet);
	}

	/*
	 * XXX if the receiver thread dies of its own accord
	 * we should do something dramatic like destroy the channel
	 * since it isn't of much use anymore...
	 */
	if (errno != EINTR && errno != EBADF) {
		struct sockaddr_in *sin;

		sin = (struct sockaddr_in *)&sdata->addr;
		printf("0x%x/%d: receive thread died: %s\n",
		       (unsigned int)ntohl(sin->sin_addr.s_addr),
		       ntohs(sin->sin_port), strerror(errno));
	}
}

#endif /* CHAN_SOCKET */
