/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <errno.h>
#include <string.h>

#include "channel.h"
#include "demux.h"
#include "demux_tlist.h"
#include "packet.h"
#include "threadpool.h"
#include "thread.h"
#include "mempool.h"
#include "misc.h"
#include "options.h"
#include "nodeos.h"

#ifdef CHAN_CLOCK
extern void chan_clock_init(void);
extern int chan_clock_create(channel_t *chan);
extern void chan_clock_destroy(channel_t *chan);
#endif
#ifdef CHAN_ETH
extern void chan_eth_init(void);
extern int chan_eth_create(channel_t *chan);
extern void chan_eth_destroy(channel_t *chan);
extern void chan_eth_send(channel_t *chan, packet_t *packet);
#endif
#ifdef CHAN_IF
extern void chan_if_init(void);
extern int chan_if_create(channel_t *chan);
extern void chan_if_destroy(channel_t *chan);
extern void chan_if_send(channel_t *chan, packet_t *packet);
#endif
#ifdef CHAN_SOCKET
extern void chan_socket_init(void);
extern int chan_socket_create(channel_t *chan);
extern void chan_socket_destroy(channel_t *chan);
extern void chan_socket_send(channel_t *chan, packet_t *packet);
#endif

/*
 * Protocol-specific processing modules.
 */
static struct chanops allchans[] =
{
#ifdef CHAN_CLOCK
{
	"clock",		chan_clock_init,
	chan_clock_create,	chan_clock_destroy,	0
},
#endif
#ifdef CHAN_ETH
{
	"eth",			chan_eth_init,
	chan_eth_create,	chan_eth_destroy,	chan_eth_send
},
#endif
#ifdef CHAN_IF
{
	"if",			chan_if_init,
	chan_if_create,		chan_if_destroy,	chan_if_send
},
#endif
#ifdef CHAN_SOCKET
{
	"socket",		chan_socket_init,
	chan_socket_create,	chan_socket_destroy,	chan_socket_send
},
#endif
{	0,	0,	0,	0,	0 }
};

void
channel_init(void)
{
	struct chanops *ops;

	for (ops = allchans; ops->name; ops++)
		if (ops->init)
			ops->init();

	if_init();
}

/*
 * Channel this thread is processing for, or zero if none.
 */
channel_t *
channel_current(void)
{
	return thread_current()->chan;
}

/*
 * ``Creates'' a channel object.
 * It acutally initializes an existing piece of memory (``chan'') to make it
 * a channel object.  The ideal here is that we allocate no memory, the
 * caller provides us with pointers to everything we need.  This includes
 * not only the channel struct, but the proto, addr, and demux structs as
 * well; i.e., anything that is passed by reference.  When the channel is
 * actually destroyed (which may be any time after channel_destroy is called
 * due to async activity in channel threads) we call the given cleanup
 * routine to free memory. 
 *
 * XXX we still allocate channel protocol-specific data and channel buffers
 * at this level.
 */
int
channel_create(channel_t *chan, int ctype, flow_t *flow,
	       channel_proto_t *proto, channel_addr_t *addr,
	       struct demux_tlist *demux, int maxthreads, int bufbytes,
	       void *qos, channel_recv_t func, void *farg,
	       void (*cleanup)(channel_t *))
{
	struct chanops *ops;
	int err;

	assert(ctype != CHAN_CUT);

	/*
	 * See if we know how to handle the desired protocol processing.
	 * XXX assumes protocol data is a string.
	 */
	for (ops = allchans; ops->name; ops++)
		if (strncmp(proto->data, ops->name, strlen(ops->name)) == 0)
			break;
	if (ops->name == 0)
		return EINVAL;

	memset(chan, 0, sizeof *chan);

	/*
	 * XXX set this now so that functions we call know who to charge
	 * for memory.  We officially link into the flow after everything
	 * has been setup.
	 */
	chan->flow = flow;

	/*
	 * XXX CPU/memory resources inherited from the flow right now
	 */
	chan->tpool = chan->flow->tpool;
	chan->mpool = flow->mpool;

	chan->maxthreads = maxthreads;
	chan->addr = addr;
	chan->proto = proto;
	chan->ops = ops;
	chan->flags = ctype | CHAN_ISRAW;
	chan->cleanup = cleanup;
	ilock_init(&chan->lock);

	/*
	 * Do channel-type specific setup
	 */
	if (ctype == CHAN_IN) {
		chan->tdata.in.demux = demux;
		chan->tdata.in.func = func;
		chan->tdata.in.farg = farg;

		/*
		 * Allocate receive buffers.
		 * Note we don't hold the lock here.
		 */
		err = channel_createbufs(chan, bufbytes);
		if (err) {
			ilock_deinit(&chan->lock);
			return err;
		}
	}

	/*
	 * Do channel-protocol specific setup
	 */
	channel_lock(chan);
	err = ops->create(chan);
	if (err) {
		if (ctype == CHAN_IN)
			channel_destroybufs(chan);
		channel_unlock(chan);
		ilock_deinit(&chan->lock);
		return err;
	}

	/*
	 * Finally link to the flow.
	 * We drop the channel lock first to adhere to our locking order.
	 */
	channel_unlock(chan);
	flow_addchannel(flow, chan);

	return 0;
}

void
channel_destroy(channel_t *chan)
{
#if 0
	int me = 0;
#endif

	assert_channel_valid(chan);

#if 0
	if (channel_current() == chan)
		me = 1;
#endif

	if (chan->flags & CHAN_WITHCUT)
		return;

	/*
	 * XXX fer now band-aid.
	 * Synchronization with channel receive interrupt routines is
	 * a mess right now.  It is bad enough that we don't hold the
	 * channel lock the entire time here, but some interrupt handlers
	 * don't lock it anyway.
	 */
	chan->flags |= CHAN_DYING;

	/*
	 * Remove from the flow before we lock the channel
	 */
	flow_removechannel(chan->flow, chan);

	channel_lock(chan);

	/*
	 * For cut channels, we destroy the constituent channels first.
	 */
	if (channel_type(chan) == CHAN_CUT) {
		channel_t *cchan;

		/*
		 * XXX flaming crock alert!
		 * Clear the WITHCUT flag so we can destroy the component
		 * channel.  To keep the receive function from flipping out,
		 * we mark the channel as DYING.
		 *
		 * I should just break down and ref count these things...
		 */
		cchan = chan->tdata.cut.inchan;
		assert(chan != cchan);
		channel_lock(cchan);
		cchan->flags |= CHAN_DYING;
		cchan->flags &= ~CHAN_WITHCUT;
		channel_unlock(cchan);
		channel_destroy(cchan);

		cchan = chan->tdata.cut.outchan;
		assert(chan != cchan);
		channel_lock(cchan);
		cchan->flags |= CHAN_DYING;
		cchan->flags &= ~CHAN_WITHCUT;
		channel_unlock(cchan);
		channel_destroy(cchan);

		channel_unlock(chan);
		ilock_deinit(&chan->lock);
		chan->cleanup(chan);
		return;
	}

#if 0
	/*
	 * Shutdown threads working for the channel
	 */
	if (chan->tpool)
		threadpool_kill(chan->tpool, chan);
#endif

	/*
	 * Channel-protocol specific shutdown
	 */
	chan->ops->destroy(chan);

	/*
	 * Channel-type specific shutdown
	 */
	if (channel_type(chan) == CHAN_IN)
		if (chan->tdata.in.rbufs)
			channel_destroybufs(chan);

	if (chan->curthreads == 0) {
		channel_unlock(chan);
		ilock_deinit(&chan->lock);
		chan->cleanup(chan);
	} else
		channel_unlock(chan);

#if 0
	/* XXX this isn't good */
	if (me)
		thread_exit();
#endif
}

int
channel_send(channel_t *chan, packet_t *packet)
{
	assert_channel_valid(chan);

	if (channel_type(chan) != CHAN_OUT)
		return EINVAL;
	if (chan->flags & CHAN_DYING)
		return EBADF; /* XXX */

	channel_lock(chan);
	chan->curthreads++;
	if (chan->ops->send)
		chan->ops->send(chan, packet);
	chan->curthreads--;
	channel_unlock(chan);

	return 0;
}

void
channel_thread_start(channel_t *chan, packet_t *packet)
{
	int err, inkern;
	void *pktarg;

	assert_channel_valid(chan);

	channel_lock(chan);

	/*
	 * XXX special case for in side of a cut-through channel.
	 *
	 * The handler function for a cut-channel is run in the NodeOS
	 * and is passed an internal packet object.
	 *
	 * The handler function for an in-channel is run in the EE and
	 * is passed the external handle for a packet.
	 */
	if (chan->flags & CHAN_WITHCUT) {
		inkern = 1;
		pktarg = packet;
	} else {
		inkern = 0;
		pktarg = NODEOS_EXTERN_PACKET(packet);
	}

	/*
	 * If the channel has threads available, fire one off.
	 */
	if (chan->curthreads < channel_maxthreads(chan)) {
		err = threadpool_thread_schedule(chan->tpool, chan, inkern,
					(thread_func_t)chan->tdata.in.func,
					pktarg, chan->tdata.in.farg);
		if (err == 0) {
			chan->curthreads++;
			channel_unlock(chan);
			return;
		}
	}
	channel_unlock(chan);
	packet_return(packet);
}

void
channel_thread_exit(channel_t *chan)
{
	assert_channel_valid(chan);

	channel_lock(chan);
	if (--chan->curthreads == 0 && (chan->flags & CHAN_DYING)) {
		channel_unlock(chan);
		ilock_deinit(&chan->lock);
		chan->cleanup(chan);
	} else
		channel_unlock(chan);
}

/*
 * Create one or more buffers an add them to a channel's list of receive
 * buffers.  Called during channel creation.
 *
 * @allocations: ``size'' bytes charged to the channel's flow.
 */
int
channel_createbufs(channel_t *chan, size_t size)
{
	struct chanbuf *fcb, *cb, *lcb;
	mempool_t *mpool;
	int nbufs;

	assert_channel_valid(chan);
	assert(channel_type(chan) == CHAN_IN);

	/*
	 * Make sure we get at least one buffer out of this
	 */
	nbufs = size / sizeof(struct chanbuf);
	if (nbufs == 0)
		return EINVAL;

	mpool = channel_mempool(chan);

	/*
	 * Allocate and chain the new bufs together.
	 * We use the first of the buffer for this.
	 */
	lcb = 0;
	for (fcb = 0; nbufs--; fcb = cb) {
		cb = mempool_alloc(mpool, sizeof *cb);
		if (cb == 0) {
			for (cb = fcb; cb != 0; cb = fcb) {
				fcb = *(struct chanbuf **)cb;
				mempool_free(mpool, cb, sizeof *cb);
			}
			return ENOMEM;
		}
		*(struct chanbuf **)cb = fcb;
		if (fcb == 0)
			lcb = cb;
	}

	/*
	 * Now add them to the channel's list.
	 * This is done is a separate step to avoid holding the channel
	 * lock too long.
	 */
	channel_lock(chan);
	*(struct chanbuf **)lcb = chan->tdata.in.rbufs;
	chan->tdata.in.rbufs = fcb;
	channel_unlock(chan);

	return 0;
}

void
channel_addbuf(channel_t *chan, struct chanbuf *cbuf)
{
	assert_channel_valid(chan);
	assert(channel_type(chan) == CHAN_IN);

	channel_lock(chan);
	*(struct chanbuf **)cbuf = chan->tdata.in.rbufs;
	chan->tdata.in.rbufs = cbuf;
	channel_unlock(chan);
}

/*
 * Remove and free all buffers associated with a channel.
 * Called during channel destroy.
 * Channel should be locked on call and return.
 *
 * @deallocation: all buffers on the channel list
 */
void
channel_destroybufs(channel_t *chan)
{
	struct chanbuf *fcb, *cb;
	mempool_t *mpool;

	assert_channel_valid(chan);
	assert(channel_type(chan) == CHAN_IN);

	/*
	 * Loop through, freeing all the buffers.
	 * Time spent holding the lock is not an issue here
	 * since the channel is dying anyway.
	 */
	mpool = channel_mempool(chan);
	for (cb = chan->tdata.in.rbufs; cb != 0; cb = fcb) {
		fcb = *(struct chanbuf **)cb;
		mempool_free(mpool, cb, sizeof *cb);
	}
}

static void
channel_packet_destroy(channel_t *chan, void *ptr, size_t size)
{
	assert_channel_valid(chan);
	mempool_free(channel_mempool(chan), ptr, size);
}

/*
 * XXX note: assumes all buffers are the same size
 */
int
channel_packet_create(channel_t *chan, void *data, int dlen, off_t offset, 
		      int mustcopy, packet_t **out_packet,
		      struct chanbuf **out_buf)
{
	struct chanbuf *newbuf;
	packet_t *packet;
	int err;

	assert_channel_valid(chan);
	assert(channel_type(chan) == CHAN_IN);

	channel_lock(chan);

	/*
	 * Channel is shutting down
	 */
	if (chan->flags & CHAN_DYING) {
		channel_unlock(chan);
		return EINVAL;
	}

	/*
	 * Ensure the channel has resources for this packet
	 */
	newbuf = chan->tdata.in.rbufs;
	if (newbuf == 0) {
		channel_unlock(chan);
		return ENOMEM;
	}
	chan->tdata.in.rbufs = *(struct chanbuf **)newbuf;
	
	/*
	 * Copy the packet if necessary
	 */
	if (mustcopy) {
		void *tmp;

		memcpy(newbuf, data, dlen);
		tmp = data;
		data = newbuf;
		newbuf = tmp;
	}

	/*
	 * Create a packet object from the buffer.
	 */
	err = packet_create_external(chan->flow, chan, data, dlen, offset,
				     channel_packet_destroy,
				     sizeof(struct chanbuf), &packet);
	if (err) {
		if (mustcopy)
			newbuf = data;
		*(struct chanbuf **)newbuf = chan->tdata.in.rbufs;
		chan->tdata.in.rbufs = newbuf;
		channel_unlock(chan);
		return err;
	}
	channel_unlock(chan);

	*out_packet = packet;
	if (out_buf)
		*out_buf = newbuf;
	return 0;
}

/*
 * Create and install a demux object on the given demux list.
 * Channel should be locked on call.
 */
int
channel_demux_add(channel_t *chan, struct demux **listp)
{
	demux_t *demux;
	int prio, err;

	assert_channel_valid(chan);
	assert(channel_type(chan) == CHAN_IN);

	if (chan->tdata.in.demux) {
		err = demux_create(chan, demux_tlist_match,
				   chan->tdata.in.demux, &demux);
		prio = chan->tdata.in.demux->ntriples; /* XXX */
	} else {
		err = demux_create(chan, 0, 0, &demux);
		prio = 0; /* XXX */
	}

	if (err == 0) {
		err = demux_add(listp, demux, prio);
		if (err)
			demux_destroy(demux);
	}

	return err;
}

void
channel_demux_remove(channel_t *chan, struct demux **listp)
{
	demux_t *demux;

	assert_channel_valid(chan);

	demux = demux_find(*listp, chan);
	if (demux) {
		demux_remove(listp, demux);
		demux_destroy(demux);
	}
}

/*
 * Cut-through channel support
 */
int
cut_channel_create(channel_t *chan, flow_t *flow, channel_proto_t *proto,
		   channel_t *inchan, channel_t *outchan,
		   channel_recv_t rfunc, void (*cleanup)(channel_t *))
{
	memset(chan, 0, sizeof *chan);

	/*
	 * XXX set this now so that functions we call know who to charge
	 * for memory.  We officially link into the flow after everything
	 * has been setup.
	 */
	chan->flow = flow;

	/*
	 * XXX CPU/memory resources inherited from the flow right now
	 */
	chan->tpool = flow->tpool;
	chan->mpool = flow->mpool;

	chan->proto = proto;
	chan->flags = CHAN_CUT;
	chan->tdata.cut.inchan = inchan;
	chan->tdata.cut.outchan = outchan;
	chan->cleanup = cleanup;
	ilock_init(&chan->lock);

	channel_lock(chan);

	/* XXX setup protocol processing */
	assert(proto == 0);

	/*
	 * Setup the inchan's recv function.
	 * Function argument is the cut channel object.
	 */
	assert(channel_type(inchan) == CHAN_IN);
	inchan->flags |= CHAN_WITHCUT;
	inchan->tdata.in.func = rfunc;
	inchan->tdata.in.farg = chan;

	/*
	 * Mark the outchan as part of the cutchan.
	 */
	assert(channel_type(outchan) == CHAN_OUT);
	outchan->flags |= CHAN_WITHCUT;

	/*
	 * Finally link to the flow.
	 * We drop the channel lock first to adhere to our locking order.
	 * Note that we remove the component in/out channels from the flow's
	 * list so they are not accessible by the EE any longer.  This is
	 * more of a hack than anything, but it simplifies the semantics
	 * a little--don't have to worry about what happens if someone
	 * tweaks the in/out channel while the cutchan is active.
	 *
	 * XXX argh!  Have to leave the channels pointing to the flow so
	 * that foo_current() calls work; i.e., for resource accounting. 
	 */
	channel_unlock(chan);
	flow_removechannel(flow, inchan);
	inchan->flow = flow;
	flow_removechannel(flow, outchan);
	outchan->flow = flow;
	flow_addchannel(flow, chan);

	return 0;
}

/*
 * Process an incoming packet on the inchan-side of a cutchan.
 * XXX note the cutchan is not locked, this could be bad.
 */
void
cut_channel_receive(an_packet_t anpacket, void *arg)
{
	channel_t *chan, *outchan;
	packet_t *packet;

	chan = arg;
	assert_channel_valid(chan);
	assert(channel_type(chan) == CHAN_CUT);

	outchan = chan->tdata.cut.outchan;
	assert_channel_valid(outchan);
	assert(channel_type(outchan) == CHAN_OUT);

	packet = (packet_t *)anpacket;
	assert(packet != 0);

	/* XXX do cutchan protocol processing */

#if 0
	{
		unsigned int *ip = packet_data(packet);

		printf("%p: %d: %#x/%#x/%#x/%#x\n", chan, packet_size(packet),
		       ntohl(ip[0]), ntohl(ip[1]), ntohl(ip[2]), ntohl(ip[3]));
	}
#endif

	(void)channel_send(outchan, packet);
	packet_return(packet);
}
