/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _CHANNEL_IMPL_H_
#define _CHANNEL_IMPL_H_

#include <nodeos/channel.h>
#include "flow.h"
#include "threadpool.h"
#include "packet.h"
#include "mempool.h"
#include "ilock.h"

/* XXX */
#define channel_addr_t an_chanaddr_t
#define channel_proto_t an_chanproto_t
#define channel_recv_t an_chanrecvfunc_t

struct chanops;
struct chanbuf;
struct demux_tlist;

struct channel {
	struct channel	 *next;	   /* link of all channels for a flow */
	flow_t		 *flow;	   /* flow we are associated with */
	threadpool_t	 *tpool;   /* thread pool for our CPU resources */
	int		 curthreads;/* threads processing packets */
	int		 maxthreads;/* max threads allowed */
	channel_addr_t	 *addr;	   /* peer addressing info */
	channel_proto_t	 *proto;   /* protocol processing info */
	struct chanops	 *ops;	   /* operations */
	mempool_t	 *mpool;   /* memory pool for our memory resources */
	int		 flags;	   /* channel type and assorted flags */
	union {
		struct {
			struct demux_tlist *demux; /* demux information */
			struct chanbuf	   *rbufs; /* receive buffers */
			channel_recv_t	   func;   /* delivery function */
			void		   *farg;  /* delivery function arg */
		} in;
		struct {
			struct channel	*inchan;
			struct channel	*outchan;
		} cut;
	}		 tdata;	   /* channel type specific data */
	void		 *data;	   /* channel-protocol specific data */
	void		 (*cleanup)(struct channel *); /* mem dealloc func */
	ilock_t		 lock;
};
typedef struct channel channel_t;

/* channel types */
#define CHAN_IN		1
#define CHAN_OUT	2
#define CHAN_CUT	3
#define CHAN_TMASK	3

/* channel flags */
#define CHAN_DYING	0x04	/* channel has been destroyed */
#define CHAN_WITHCUT	0x08	/* associated with a cut-through channel */
#define CHAN_ISRAW	0x10	/* provide protocol headers (if possible) */

#define channel_lock(c)		ilock_lock(&(c)->lock)
#define channel_unlock(c)	ilock_unlock(&(c)->lock)
#define channel_islocked(c)	ilock_islocked(&(c)->lock)

#define channel_maxthreads(c)	((c)->maxthreads)
#define channel_name(c)		((char *)((c)->proto->data))
#define channel_namelen(c)	((c)->proto->len)
#define channel_type(c)		((c)->flags & CHAN_TMASK)
#define channel_mempool(c)	((c)->mpool)

#ifdef MEMDEBUG
#define assert_channel_valid(c) \
	assert((c) != 0 && (c) != (void *)0xdddddddd), \
	assert(*(unsigned int *)(c) != 0xdddddddd)
#else
#define assert_channel_valid(c)	assert((c) != 0)
#endif

struct chanops {
	char *name;
	void (*init)(void);
	int (*create)(channel_t *);
	void (*destroy)(channel_t *);
	void (*send)(channel_t *, packet_t *);
};

/*
 * In-channel receive buffers.
 */
struct chanbuf {
	char data[NODEOS_CHAN_BUFSIZE];
};

struct demux;

void channel_init(void);
channel_t *channel_current(void);
int channel_create(channel_t *chan, int ctype, flow_t *flow,
		   channel_proto_t *proto, channel_addr_t *addr,
		   struct demux_tlist *demux,
		   int maxthreads, int bufbytes, void *qos,
		   channel_recv_t func, void *farg,
		   void (*cleanup)(channel_t *));
void channel_destroy(channel_t *chan);
int channel_send(channel_t *chan, packet_t *packet);
void channel_thread_start(channel_t *chan, packet_t *packet);
void channel_thread_exit(channel_t *chan);
int channel_packet_create(channel_t *chan, void *data, int dlen,
			  off_t offset, int mustcopy,
			  packet_t **out_packet, struct chanbuf **out_buf);
int channel_createbufs(channel_t *chan, size_t size);
void channel_addbuf(channel_t *chan, struct chanbuf *cbuf);
void channel_destroybufs(channel_t *chan);
int channel_demux_add(channel_t *chan, struct demux **listp);
void channel_demux_remove(channel_t *chan, struct demux **listp);

int cut_channel_create(channel_t *chan, flow_t *flow, channel_proto_t *proto,
		       channel_t *inchan, channel_t *outchan,
		       channel_recv_t rfunc, void (*cleanup)(channel_t *));
void cut_channel_receive(an_packet_t packet, void *arg);

#endif /* _CHANNEL_IMPL_H_ */
