/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <pthread.h>
#include <errno.h>
#include <assert.h>
#include <sys/time.h>

#include "os.h"
#include "options.h"

#ifdef CHAN_ETH
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>


#include <net/if.h>
#include <netinet/in.h>

#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <netinet/if_ether.h>

#include <sys/uio.h>
#endif /* CHAN_ETH */

void
os_init_basic(void)
{
}

void
os_init_mem(void)
{
}

void
os_init_devices(void)
{
}

/*
 * Set the priority of the given thread.
 * This routine should map standard NodeOS priority values into
 * implementation specific priorities.
 */
int
os_sched_setprio(struct thread *thread, int nprio)
{
	return 0;
}

/*
 * This function is called by the assert() macro defined in assert.h;
 * it's also a nice simple general-purpose panic function.
 */
void
panic(const char *fmt, ...)
{
	va_list vl;

	va_start(vl, fmt);
	vprintf(fmt, vl);
	va_end(vl);

	printf("\n");

	exit(1);
}


static void
sigcontext_dump_stack_trace(struct sigcontext *scp)
{
	unsigned int *fp, i;

	fp = (unsigned int *)scp->ebp;

	printf("Backtrace:");
	for (i = 0; i < 16; i++) {
		if ((i % 8) == 0)
			printf("\n");

		fp = (int *)(*fp);
		if (!(*(fp + 1) && *fp))
			break;
		printf(" %08x", *(fp + 1));
	}
	printf("\n");
}

void
os_dump_state(void *state)
{
	struct sigcontext *scp = state;
	static int indump;

	if (indump++)
		return;

	printf("EAX %08lx EBX %08lx ECX %08lx EDX %08lx\n",
	       scp->eax, scp->ebx, scp->ecx, scp->edx);
	printf("ESI %08lx EDI %08lx EBP %08lx ESP %08lx\n",
	       scp->esi, scp->edi, scp->ebp, scp->esp);
	printf("EIP %08lx EFLAGS %08lx\n", scp->eip, scp->eflags);
	printf("CS %04x SS %04x DS %04x ES %04x\n",
	       scp->cs & 0xffff, scp->ss & 0xffff,
	       scp->ds & 0xffff, scp->es & 0xffff);
	sigcontext_dump_stack_trace(scp);
	indump = 0;
}

#ifdef CHAN_ETH
int
os_open_eth(char *dev)
{
	int fd;
	struct ifreq ifr;
	struct sockaddr_ll sa;
	struct packet_mreq mr;

	printf("opening device %s\n", dev);
	
	fd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (fd < 0) {
		/* print error */
		return -1;
	}

	memset(&sa, 0, sizeof(sa));
	strcpy(ifr.ifr_name, dev);

	if (ioctl(fd, SIOCGIFINDEX, &ifr) < 0) {
		/* print error */
		close(fd);
		return -1;
	}
	
	sa.sll_family = AF_PACKET;
	sa.sll_ifindex = ifr.ifr_ifindex;
	sa.sll_protocol = htons(ETH_P_ALL);

	if (bind(fd, (struct sockaddr *)&sa, sizeof(sa))) {
		/* print error */
		close(fd);
		return -1;
	}

	memset(&mr, 0, sizeof(mr));
	mr.mr_ifindex = sa.sll_ifindex;
	mr.mr_type = PACKET_MR_PROMISC;

	if (setsockopt(fd, SOL_PACKET, PACKET_ADD_MEMBERSHIP, (char *)&mr,
		       sizeof(mr)) < 0)
	{
		/* print error */
		close(fd);
		return -1;
	}

	return fd;
}

int
os_read_packet(int fd, char *buf, int bufsize)
{
	struct sockaddr_ll from;
	int fromlen;
	ssize_t count;

	
  again:		
	fromlen = sizeof(from);
	count = recvfrom(fd, buf, bufsize, 0,
			 (struct sockaddr *)&from, &fromlen);

	if (count < 0 && errno == EINTR)
		goto again;

	return count;
}

int
os_write_packet(int fd, char *buf, int bufsize, struct ether_header *eh)
{
	struct iovec msg_iov[2];
	struct msghdr msg;

	memset(&msg, 0, sizeof(msg));
	msg.msg_iov = (struct iovec *)&msg_iov;

	if (eh != 0) {
		msg_iov[0].iov_base = eh;
		msg_iov[0].iov_len  = ETHER_HDR_LEN;
		msg_iov[1].iov_base = buf;
		msg_iov[1].iov_len  = bufsize;
		msg.msg_iovlen = 2;
	} else {
		msg_iov[0].iov_base = buf;
		msg_iov[0].iov_len  = bufsize;
		msg.msg_iovlen = 1;
	}
	
	return (int)sendmsg(fd, &msg, 0);
}
#endif /* CHAN_ETH */
