/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Hodge-podge of stuff for OSKit
 */
#include <stdio.h>
#include <errno.h>

#include <oskit/clientos.h>
#include <oskit/startup.h>
#include <oskit/threads/pthread.h>

#include <oskit/dev/dev.h>
#include <oskit/dev/osenv.h>
#include <oskit/dev/blk.h>
#include <oskit/dev/ethernet.h>
#include <oskit/dev/bus.h>
#include <oskit/dev/linux.h>

#include "os.h"
#include "options.h"

#include <nodeos/thread.h>
#include "thread.h"

#ifdef GPROF
void oskit_init_gprof(void);
#endif
#ifdef HAVE_DEVICES
static int verbose = 1;
#endif

/*
 * Get the basic OSKit ``stuff'' up and running.
 */
void
os_init_basic(void)
{
	oskit_clientos_init_pthreads();
	start_clock();
	start_pthreads();
#ifdef GPROF
	start_fs_bmod();
	oskit_init_gprof();
#endif
}

#ifdef GPROF

static char *_gprof_args[] = {"gprof", "nodeos_p", 0};
static int _gprof_hz = 8192;

/*
 * Equiv of startup/start_gprof.
 * We set the sample rate and use our own cleanup so we can call gprof
 * with arguments.
 */
void
oskit_init_gprof(void)
{
	extern int enable_gprof, gprof_hz;
	void oskit_dump_gprof(void *);
	void _mcleanup(void *);

	if (!enable_gprof) {
		printf("start_gprof: GPROF is not enabled in kern library!");
		return;
	}
	
	gprof_hz = _gprof_hz;
	startup_atexit(oskit_dump_gprof, NULL);
	startup_atexit(_mcleanup, NULL);
}

void
oskit_start_gprof(void)
{
	extern unsigned long _start, etext;
	void monstartup(unsigned long lowpc, unsigned long highpc);

	printf("Starting profile of `%s' ([%p-%p]), %d samples/sec\n",
	       _gprof_args[1], &_start, &etext, _gprof_hz);
	monstartup((unsigned long)&_start, (unsigned long)&etext);
}

void
oskit_stop_gprof(void)
{
	void moncontrol(int mode);

	moncontrol(0);
	printf("Profile done\n");
}

void oskit_dump_gprof(void *arg)
{
	gprof(2, _gprof_args);
}

/*
 * Override this to prevent it from happening out of multiboot_main.
 */
void
base_gprof_init()
{
}

#endif

/*
 * Initialize and probe device drivers.
 */
void
os_init_devices(void)
{
#ifdef HAVE_DEVICES
	osenv_process_lock();
	start_devices();
	osenv_process_unlock();
#endif
}

#ifdef HAVE_DEVICES
/*
 * This is our own version of start devices so we can be more selective.
 * Comment this out, if you want all the enet and disk devices.
 */
void
start_devices(void)
{
	static void init_net(void);
	static void init_disk(void);
	oskit_osenv_t *osenv;
	static int run;
	
	if (run++)
		return;

	printf("Initializing device drivers...\n");
	osenv = start_osenv();
	oskit_dev_init(osenv);
	oskit_linux_init_osenv(osenv);

#ifdef HAVE_NET
	init_net();
#endif
#ifdef HAVE_DISK
	init_disk();
#endif

	if (verbose)
		oskit_dump_drivers();

	printf("Probing devices...\n");
	oskit_dev_probe();

	if (verbose)
		oskit_dump_devices();
}

#ifdef HAVE_NET
/*
 * Equiv of oskit_linux_init_ethernet
 *
 * Insert individual drivers you want or call oskit_linux_init_ethernet
 * for all of them.
 */
static void
init_net(void)
{
#ifdef i386
	oskit_linux_init_ethernet_tulip();
	oskit_linux_init_ethernet_eepro100();
#endif
#ifdef arm32
	oskit_linux_init_ethernet_cs89x0();
#endif
}

/*
 * XXX override startup library.
 *
 * This routine would be pulled in by startup_network_router and the
 * default version would pull in all enet drivers.
 */
void
start_net_devices(void)
{
}

#endif

#ifdef HAVE_DISK
/*
 * Equiv of oskit_linux_init_blk
 *
 * Insert individual drivers you want or call oskit_linux_init_blk
 * for all of them.
 */
static void
init_disk(void)
{
	oskit_linux_init_ide();
	oskit_linux_init_scsi_ncr53c8xx();
	oskit_linux_init_scsi_aic7xxx();
}
#else

/*
 * XXX guess what?  A hack to keep from dragging in a ton o' crap!
 * The default Linux device_setup pulls in all the block device code:
 * ide, scsi, you name it.  Turning on interrupts seems to be the only
 * thing we need if we don't care about disk devices.
 */
void
FDEV_LINUX_device_setup(void)
{
	osenv_intr_enable();
}
#endif
#endif

/*
 * Set the priority of the given thread.
 * This routine should map standard NodeOS priority values into
 * implementation specific priorities.
 */
int
os_sched_setprio(thread_t *thread, int nprio)
{
	int prio, err;

	switch (nprio) {
	case AN_THREAD_PRIO_IDLE:
		prio = PRIORITY_MIN;
		break;
	case AN_THREAD_PRIO_EXCLUSIVE:
		prio = PRIORITY_MAX;
		break;
	case AN_THREAD_PRIO_NORMAL:
		prio = PRIORITY_NORMAL;
		break;
	default:
		return EINVAL;
	}
	err = oskit_pthread_setprio(thread->id, prio);
	if (err == 0)
		thread->pprio.priority = prio;

	return err;
}

void
os_dump_state(void *state)
{
	sigcontext_dump((struct sigcontext *)state);
}
