/*
 * Copyright (c) 1999, 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The One True Memory Object (TOTMO).
 *
 * In theory, every memory allocation from anywhere in the OSKit channels
 * through here.  For any such call, we will charge the memory to the current
 * flow.  In doing so, we are hoping that the matching deallocation will
 * be in the context of the same flow!
 */

#include <assert.h>
#include <oskit/com.h>
#include <oskit/com/mem.h>
#include <oskit/com/services.h>
#include <oskit/lmm.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>			/* mem_lock - BOGUS! */
#include <oskit/machine/phys_lmm.h>

#ifdef MEMDEBUG
#include <oskit/memdebug.h>
int nodeos_memdebug_on;
#endif

#define MEMLOGGING

/* A common symbol that is overridden by the kernel library */
lmm_t malloc_lmm;

/* the COM object */
static struct oskit_mem_ops nodeos_mem_ops;
static oskit_mem_t nodeos_mem  = { &nodeos_mem_ops };

/* there is no more core */
#define MORECORE(size, flags) ({ \
	panic("NodeOS: out of memory"); \
}) 


#ifdef MEMLOGGING

#define MEMLOGSIZE	(64*1024)	/* # of entries */
#define MEMLOGTSIZE	13		/* size of stack trace to keep */

struct memlog {
	int event;
	void *addr;
	oskit_size_t size;
	unsigned int trace[MEMLOGTSIZE];
} memlog[MEMLOGSIZE], *memlogptr, *ememlogptr;

#define MEMLOG(e, a, s) \
if (memlogging) { \
	memlogptr->event = (e); \
	memlogptr->addr = (a); \
	memlogptr->size = (s); \
	store_stack_trace(memlogptr->trace, MEMLOGTSIZE); \
	if (++memlogptr == ememlogptr) \
		memlogptr = memlog; \
}

void store_stack_trace(unsigned *backtrace, int max_len);
void dump_stored_stack_trace(int (*__printf)(const char *,...), 
			     unsigned *backtrace, int max_st_levels);
void memlog_dump(void);

#else
#define MEMLOG(e, a, s)
#endif
int memlogging;

#define EV_ALLOC	1
#define EV_FREE		2


void
os_init_mem(void)
{
	/*
	 * XXX oskit_mem_init should have already been called by the
	 * low-level OSKit startup, but we call it again anyway.
	 */
	(void)oskit_mem_init();

#ifdef MEMLOGGING
	/*
	 * Start logging memory events.
	 */
	memlogptr = memlog;
	ememlogptr = &memlog[MEMLOGSIZE];
#if 0
	atexit(memlog_dump);
	memlogging = 1;
#endif
#endif
#ifdef MEMDEBUG
	/*
	 * Start monitoring memory allocation.
	 */
	nodeos_memdebug_on = 1;
#endif
}

static OSKIT_COMDECL 
nodeos_mem_query(oskit_mem_t *m, const oskit_iid_t *iid, void **out_ihandle)
{ 
        if (memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
            memcmp(iid, &oskit_mem_iid, sizeof(*iid)) == 0) {
                *out_ihandle = m;
                return 0; 
        }
  
        *out_ihandle = 0;
        return OSKIT_E_NOINTERFACE;
};

static OSKIT_COMDECL_U
nodeos_mem_addref(oskit_mem_t *m)
{
	/* No reference counting; only one of them */
	return 1;
}

static OSKIT_COMDECL_U
nodeos_mem_release(oskit_mem_t *m)
{
	/* No reference counting; only one of them */
	return 1;
}

static void * OSKIT_COMCALL
nodeos_mem_alloc(oskit_mem_t *m, oskit_u32_t size, oskit_u32_t flags)
{
	lmm_flags_t	lmm_flags = 0;
	oskit_u32_t	*chunk;

	if (flags & OSKIT_MEM_ISADMA_MEM)
		lmm_flags |= LMMF_16MB;

#ifdef MEMDEBUG
	/*
	 * XXX memdebug cannot handle any LMM flags
	 */
	if (nodeos_memdebug_on && lmm_flags == 0) {
		if (flags & OSKIT_MEM_AUTO_SIZE)
			chunk = memdebug_traced_malloc(size, 0,
						       __FILE__, __LINE__);
		else
			chunk = memdebug_traced_smalloc(size, 0,
							__FILE__, __LINE__);
		return chunk;
	}
#endif

	mem_lock();
	if (flags & OSKIT_MEM_AUTO_SIZE) {
		size += sizeof(oskit_u32_t);

		while (!(chunk = lmm_alloc(&malloc_lmm, size, lmm_flags))) {
			MORECORE(size, lmm_flags);
		}
		MEMLOG(EV_ALLOC, chunk, size);

		*chunk++ = size;
	}
	else {
		while (!(chunk = lmm_alloc(&malloc_lmm, size, lmm_flags))) {
			MORECORE(size, lmm_flags);
		}
		MEMLOG(EV_ALLOC, chunk, size);
	}
	mem_unlock();

	return chunk;
}

static void * OSKIT_COMCALL
nodeos_mem_realloc(oskit_mem_t *m, void *ptr,
	    oskit_u32_t oldsize, oskit_u32_t newsize, oskit_u32_t flags)
{
	lmm_flags_t	lmm_flags = 0;
	oskit_u32_t	*chunk;

	if (flags & OSKIT_MEM_ISADMA_MEM)
		lmm_flags |= LMMF_16MB;

#ifdef MEMDEBUG
	/*
	 * XXX memdebug cannot handle any LMM flags
	 */
	if (nodeos_memdebug_on && lmm_flags == 0) {
		if (flags & OSKIT_MEM_AUTO_SIZE)
			chunk = memdebug_traced_malloc(newsize, 0,
						       __FILE__, __LINE__);
		else
			chunk = memdebug_traced_smalloc(newsize, 0,
							__FILE__, __LINE__);
		if (chunk) {
			memcpy(chunk, ptr,
			       oldsize < newsize ? oldsize : newsize);
			if (flags & OSKIT_MEM_AUTO_SIZE)
				memdebug_traced_free(ptr,
						     __FILE__, __LINE__);
			else
				memdebug_traced_sfree(ptr, oldsize,
						      __FILE__, __LINE__);
		}
		return chunk;
	}
#endif

	mem_lock();
	if (flags & OSKIT_MEM_AUTO_SIZE) {
		oskit_size_t	*op;

		op = (oskit_size_t *)ptr;
		oldsize  = *--op;
		newsize += sizeof(oskit_u32_t);

		while (!(chunk = lmm_alloc(&malloc_lmm, newsize, lmm_flags))) {
			MORECORE(newsize, lmm_flags);
		}
		MEMLOG(EV_ALLOC, chunk, newsize);

		memcpy(chunk, op, oldsize < newsize ? oldsize : newsize);
		MEMLOG(EV_FREE, op, oldsize);
		lmm_free(&malloc_lmm, op, oldsize);

		*chunk++ = newsize;
	}
	else {
		while (!(chunk = lmm_alloc(&malloc_lmm, newsize, lmm_flags))) {
			MORECORE(newsize, lmm_flags);
		}
		MEMLOG(EV_ALLOC, chunk, newsize);

		memcpy(chunk, ptr, oldsize < newsize ? oldsize : newsize);
		MEMLOG(EV_FREE, ptr, oldsize);
		lmm_free(&malloc_lmm, ptr, oldsize);
	}
	mem_unlock();

	return chunk;
}

static void *OSKIT_COMCALL
nodeos_mem_alloc_aligned(oskit_mem_t *m,
		  oskit_u32_t size, oskit_u32_t flags, oskit_u32_t align)
{
	lmm_flags_t	lmm_flags = 0;
	oskit_u32_t	*chunk;
	unsigned	shift;

	if (flags & OSKIT_MEM_ISADMA_MEM)
		lmm_flags |= LMMF_16MB;

#ifdef MEMDEBUG
	/*
	 * XXX memdebug cannot handle any LMM flags
	 */
	if (nodeos_memdebug_on && lmm_flags == 0) {
		if (flags & OSKIT_MEM_AUTO_SIZE)
			chunk = memdebug_traced_memalign(align, size, 0,
							 __FILE__, __LINE__);
		else
			chunk = memdebug_traced_smemalign(align, size, 0,
							  __FILE__, __LINE__);
		return chunk;
	}
#endif

	/* Find the alignment shift in bits.  XXX use proc_ops.h  */
	for (shift = 0; (1 << shift) < align; shift++);

	mem_lock();
	if (flags & OSKIT_MEM_AUTO_SIZE) {
		size += sizeof(oskit_u32_t);

		while (!(chunk =
			 lmm_alloc_aligned(&malloc_lmm, size, lmm_flags, shift,
					(1 << shift) - sizeof(oskit_u32_t)))) {
			MORECORE(size * 2, lmm_flags);
		}
		MEMLOG(EV_ALLOC, chunk, size);
		
		*chunk++ = size;
	}
	else {
		while (!(chunk = lmm_alloc_aligned(&malloc_lmm,
						size, lmm_flags, shift, 0))) {
			MORECORE(size * 2, lmm_flags);
		}
		MEMLOG(EV_ALLOC, chunk, size);
	}
	mem_unlock();

	return chunk;
}

static void OSKIT_COMCALL
nodeos_mem_free(oskit_mem_t *m, void *ptr, oskit_u32_t size, oskit_u32_t flags)
{
	/* Posix says free of NULL does nothing */
	if (ptr == 0)
		return;

#ifdef MEMDEBUG
	/*
	 * XXX linux glue does't (and can't) pass ISADMA_MEM flag to us
	 * so we detect it manually!
	 */
	if (nodeos_memdebug_on && ptr >= 0x1000000) {
		if (flags & OSKIT_MEM_AUTO_SIZE)
			memdebug_traced_free(ptr, __FILE__, __LINE__);
		else
			memdebug_traced_sfree(ptr, size, __FILE__, __LINE__);
		return;
	}
#endif

	mem_lock();
	if (flags & OSKIT_MEM_AUTO_SIZE) {
		oskit_u32_t *chunk = (oskit_u32_t *)ptr - 1;

		MEMLOG(EV_FREE, chunk, *chunk);
		lmm_free(&malloc_lmm, chunk, *chunk);
	}
	else {
		MEMLOG(EV_FREE, ptr, size);
		lmm_free(&malloc_lmm, ptr, size);
	}
	mem_unlock();
}

static oskit_u32_t OSKIT_COMCALL
nodeos_mem_getsize(oskit_mem_t *m, void *ptr)
{
#ifdef MEMDEBUG
	if (nodeos_memdebug_on) {
		/* XXX no public interface for this */
		int memdebug_getsize(void *ptr, int *sz, char *file, int line);
		oskit_u32_t size;

		return memdebug_getsize(ptr, &size, __FILE__, __LINE__) ?
			0 : size;
	}
#endif
	if (ptr) {
		oskit_u32_t *chunk = (oskit_u32_t *)ptr - 1;

		return *chunk;
	}
	return 0;
}

static void *OSKIT_COMCALL
nodeos_mem_alloc_gen(oskit_mem_t *m,
	      oskit_u32_t size, oskit_u32_t flags,
	      oskit_u32_t align_bits, oskit_u32_t align_ofs)
{
	lmm_flags_t	lmm_flags = 0;
	oskit_u32_t	*chunk;

	if (flags & OSKIT_MEM_ISADMA_MEM)
		lmm_flags |= LMMF_16MB;

	/*
	 * Not really all that general. For now, support only non
	 * autosize allocations. 
	 */
	if (flags & OSKIT_MEM_AUTO_SIZE)
		return NULL;

#ifdef MEMDEBUG
	/*
	 * XXX memdebug doesn't support LMM flags or alignment offsets
	 */
	if (nodeos_memdebug_on && lmm_flags == 0 && align_ofs == 0) {
		chunk = memdebug_traced_smemalign((1 << align_bits), size, 0,
						  __FILE__, __LINE__);
		return chunk;
	}
#endif

	mem_lock();
	while (!(chunk = lmm_alloc_gen(&malloc_lmm, size, lmm_flags,
				       align_bits, align_ofs, 0, -1))) {
		MORECORE(size * 2, flags);
	}
	MEMLOG(EV_ALLOC, chunk, size);
	mem_unlock();

	return chunk;
}

static oskit_size_t OSKIT_COMCALL
nodeos_mem_avail(oskit_mem_t *m, oskit_u32_t flags)
{
	lmm_flags_t	lmm_flags = 0;
	oskit_size_t	avail;

	if (flags & OSKIT_MEM_ISADMA_MEM)
		lmm_flags |= LMMF_16MB;

	if (flags & OSKIT_MEM_X861MB_MEM)
		lmm_flags |= LMMF_1MB;

	mem_lock();
	avail = lmm_avail(&malloc_lmm, lmm_flags);
	mem_unlock();

	return avail;
}

static OSKIT_COMDECL_V
nodeos_mem_dump(oskit_mem_t *m)
{
	mem_lock();
	lmm_dump(&malloc_lmm);
	mem_unlock();
}

static struct oskit_mem_ops nodeos_mem_ops = {
	nodeos_mem_query,
	nodeos_mem_addref,
	nodeos_mem_release,
	nodeos_mem_alloc,
	nodeos_mem_realloc,
	nodeos_mem_alloc_aligned,
	nodeos_mem_free,
	nodeos_mem_getsize,
	nodeos_mem_alloc_gen,
	nodeos_mem_avail,
	nodeos_mem_dump,
};

/*
 * Not a lot to do in this impl.
 */
oskit_mem_t *
oskit_mem_init(void)
{
	static int called;

	if (called == 0) {
		called++;
	}
	return &nodeos_mem;
}

void
memlog_dump(void)
{
#ifdef MEMLOGGING
	struct memlog *aptr, *ptr;
	unsigned int alloced = 0, freed = 0;
	int events, matches;

	memlogging = 0;

	/*
	 * Pass 1, cancel out matching alloc/free
	 */
	printf("Memory log: ");
	ptr = memlogptr;
	events = matches = 0;
	do {
		if (ptr->event != 0)
			events++;
		if (ptr->event == EV_ALLOC) {
			aptr = ptr;
			do {
				if (ptr->event == EV_FREE &&
				    ptr->addr == aptr->addr &&
				    ptr->size == aptr->size) {
					aptr->event = ptr->event = 0;
					ptr = aptr;
					events++;
					matches++;
				} else if (++ptr == ememlogptr)
					ptr = memlog;
			} while (aptr != ptr);
		}
		if (++ptr == ememlogptr)
			ptr = memlog;
	} while (ptr != memlogptr);
	printf("%d events, %d matching alloc/frees, ", events, matches);

	/*
	 * Pass 2, dump
	 */
	printf("%d unmatched alloc/frees:\n", events - (2*matches));
	do {
		if (ptr->event != 0) {
			printf("%c: %p/%d",
			       ptr->event == EV_ALLOC ? 'A' : 'F',
			       ptr->addr, ptr->size);

			printf(": ");
			dump_stored_stack_trace(printf, ptr->trace,
						MEMLOGTSIZE);
			printf("\n");

			if (ptr->event == EV_ALLOC)
				alloced += ptr->size;
			else
				freed += ptr->size;
		}
		if (++ptr == ememlogptr)
			ptr = memlog;
	} while (ptr != memlogptr);
	if (alloced)
		printf("%d bytes allocated but not freed\n", alloced);
	if (freed)
		printf("%d bytes freed but not allocated\n", freed);
#endif
	lmm_stats(&malloc_lmm);
}

#ifdef MEMDEBUG

/*
 * Routines used by memdebug to actually alloc/free memory.
 */
void *
memdebug_untraced_alloc(oskit_u32_t size,
			oskit_u32_t align_bits, oskit_u32_t align_ofs)
{
	oskit_u32_t	*chunk;

	mem_lock();
	while (!(chunk = lmm_alloc_aligned(&malloc_lmm, size, 0,
					   align_bits, align_ofs))) {
		MORECORE(size * 2, 0);
	}
	MEMLOG(EV_ALLOC, chunk, size);
	mem_unlock();

	return chunk;
}

void
memdebug_untraced_free(void *ptr, oskit_u32_t size)
{
	/* Posix says free of NULL does nothing */
	if (ptr == 0)
		return;

	mem_lock();
	MEMLOG(EV_FREE, ptr, size);
	lmm_free(&malloc_lmm, ptr, size);
	mem_unlock();
}

#endif
