/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <pthread.h>
#include <sched.h>

#include <nodeos/thread.h>

#include "nodeos.h"
#include "os.h"
#include "flow.h"
#include "thread.h"
#include "threadpool.h"
#include "channel.h"
#include "mem.h"
#include "misc.h"

//#define VERBOSE

static pthread_key_t nodeos_thread;	/* thread struct for this pthread */
static pthread_key_t nodeos_threaddata;	/* nodeos exported thread data */

/*
 * For hack thread_sleep implementation.
 * Cond is never signalled, just allows us to do a timed wait.
 */
static pthread_mutex_t sleepmutex;
static pthread_cond_t sleepcond;

static void thread_done(thread_t *thread);
static void thread_cleanup(void *arg);
static void *thread_loop(void *arg);
static void thread_intr_init(void);

void
thread_init(void)
{
	int err;

	/*
	 * Establish a thread_specific data key to hold the thread_t
	 * associated with a pthread.  This is used to determine the
	 * current thread.  Init to zero for this primortial thread.
	 */
	err = pthread_key_create(&nodeos_thread, 0);
	assert(err == 0);
	err = pthread_setspecific(nodeos_thread, 0);
	assert(err == 0);

	/*
	 * Also need a key to use for the exported thread_setdata
	 * interface.
	 */
	err = pthread_key_create(&nodeos_threaddata, 0);
	assert(err == 0);
	err = pthread_setspecific(nodeos_threaddata, 0);
	assert(err == 0);

	/*
	 * For thread sleep
	 */
	err = pthread_mutex_init(&sleepmutex, 0);
	assert(err == 0);
	err = pthread_cond_init(&sleepcond, 0);
	assert(err == 0);

	thread_intr_init();
}

thread_t *
thread_current(void)
{
	/*
	 * In the current implementation has no control over scheduling
	 * of threads, so we have to query the pthread state.
	 */
	return pthread_getspecific(nodeos_thread);
}

/*
 * Called during thread pool initialization to ensure that thread parameters
 * are compatible with the underlying thread implementation.
 */
int
thread_checkattrs(cpuspec_t *cspec)
{
	pthread_attr_t attr;
	int err;

	err = pthread_attr_init(&attr);
	assert(err == 0);
	err = pthread_attr_setstacksize(&attr, cspec->stacksize);
	return err;
}

int
thread_spawn(threadpool_t *tpool, thread_t **out_thread)
{
	thread_t *thread;
	pthread_attr_t attr;
	int err;

	assert(threadpool_islocked(tpool));

	if (tpool->curthreads >= tpool->params.max)
		return EAGAIN;

	thread = mem_alloc(sizeof *thread);
	if (thread == 0)
		return ENOMEM;
	memset(thread, 0, sizeof *thread);

	/*
	 * Set up the attributes for the thread.
	 * Currently we only set the stack size.
	 * (XXX and global scope for the benefit of Solaris)
	 */
	err = pthread_attr_init(&attr);
	assert(err == 0);
	err = pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	assert(err == 0);
	err = pthread_attr_setstacksize(&attr, tpool->params.stacksize);
	assert(err == 0);

	/*
	 * Create the pthread, it will block immediately on the pool lock.
	 */
	thread->pool = tpool;
	err = pthread_create(&thread->id, &attr, thread_loop, thread);
	if (err) {
		mem_free(thread, sizeof *thread);
		return err;
	}
	err = pthread_detach(thread->id);
	assert(err == 0);

	/* Set the default scheduling priority */
	thread->nprio = AN_THREAD_PRIO_NORMAL;
	err = os_sched_setprio(thread, thread->nprio);
	assert(err == 0);

	/* We are committed, finish up */
	ilock_init(&thread->lock);
	thread->flags = THREAD_INKERNEL;
	thread->pool = tpool;
	icond_init(&thread->waiting);
	icond_init(&thread->waitedon);
	thread->next = tpool->threads;
	tpool->threads = thread;
	tpool->curthreads++;

	if (out_thread)
		*out_thread = thread;

	return 0;
}

/*
 * All NodeOS threads spend their lives here with brief forays into EE/AA code.
 */
static void *
thread_loop(void *arg)
{
	thread_t *thread = arg;
	threadpool_t *tpool;
	struct threadwork *w, wtmp;

	/*
	 * Record our NodeOS identity in the pthread so thread_current works.
	 */
	assert(thread != 0);
	pthread_setspecific(nodeos_thread, thread);

	/*
	 * Get the threadpool lock.
	 * We will block here until our creator is finished initializing us.
	 */
	tpool = thread->pool;
	assert(tpool != 0);
	threadpool_lock(tpool);
#ifdef VERBOSE
printf("TP%p: thread %p started\n", tpool, thread);
#endif

	/*
	 * Now that we are official, push a cleanup handler, and go to work.
	 */
	pthread_cleanup_push(thread_cleanup, thread);
	while (1) {
		/*
		 * If there is no work to do we either block waiting for
		 * more or die if there are enough threads in reserve.
		 */
		while (thread->work == 0 && tpool->workhead == 0) {
			if (tpool->freethreads >= tpool->reserve)
				goto die;
			thread->freenext = tpool->freelist;
			tpool->freelist = thread;
			tpool->freethreads++;
#ifdef THREADPOOL_STATS
			if (tpool->freethreads > tpool->stats.maxcached)
				tpool->stats.maxcached = tpool->freethreads;
#endif
			if (tpool->flags & TPOOL_WAITING) {
				tpool->flags &= ~TPOOL_WAITING;
				icond_broadcast(&tpool->waiting);
			}
#ifdef VERBOSE
printf("TP%p: thread %p sleeping, freethreads=%d\n", tpool, thread, tpool->freethreads);
#endif
			icond_wait(&thread->waiting, &tpool->lock);
#ifdef VERBOSE
printf("TP%p: thread %p awake, freethreads=%d\n", tpool, thread, tpool->freethreads);
#endif
		}
		assert(thread->freenext == 0);

#ifdef THREADPOOL_STATS
		{
			int active = tpool->curthreads - tpool->freethreads;

			if (active > tpool->stats.maxactive)
				tpool->stats.maxactive = active;
		}
#endif

		/*
		 * First look for worked specifically targetted to this thread.
		 * Otherwise, grab the first item off the list.
		 */
		w = thread->work;
		if (w == 0) {
			w = tpool->workhead;
			if (w->next == 0) {
				assert(tpool->worktail == w);
				tpool->workhead = tpool->worktail = 0;
			} else
				tpool->workhead = w->next;
			tpool->worklen--;
		} else
			thread->work = 0;

		/*
		 * Make a local copy so we can free the heap element
		 */
		wtmp = *w;
		mempool_free(tpool->flow->mpool, w, sizeof *w);
		w = &wtmp;

		thread->chan = w->chan;

#ifdef VERBOSE
printf("TP%p: thread %p doing %p(%p)\n", tpool, thread, w->func, w->farg);
#endif
		/*
		 * Run the desired function.
		 *
		 * The sigsetjmp catches threads that call thread_exit
		 * explicitly (as opposed to returning from the registered
		 * upcall func).  Those threads will siglongjmp back to
		 * here unwinding their stacks.
		 *
		 * The cleanup handler ensure that anybody awaiting this
		 * thread is properly awakened in the event of a cancellation.
		 * Note the "1" argument to cleanup_pop that causes
		 * thread_done to be called even when no cancellation occured.
		 */
		pthread_cleanup_push((void (*)())thread_done, thread);
		threadpool_unlock(tpool);
		if (sigsetjmp(thread->jmpbuf, 0) == 0) {
			if (w->inkern == 0) {
				thread->flags |= THREAD_INTSON;
				NODEOS_LEAVE;
				w->func(w->farg[0], w->farg[1]);
				NODEOS_ENTER;
			} else
				w->func(w->farg[0], w->farg[1]);
		}
		tpool = thread->pool;
		pthread_cleanup_pop(1);
		threadpool_lock(tpool);
#ifdef VERBOSE
printf("TP%p: thread %p did %p(%p)\n", tpool, thread, w->func, w->farg);
#endif
	}
 die:
#ifdef VERBOSE
printf("TP%p: thread %p dying, freethreads=%d\n", tpool, thread, tpool->freethreads);
#endif
	threadpool_unlock(thread->pool);

	/*
	 * Call pthread_exit explicitly here instead of returning
	 * since the cleanup handler must be "in scope" when we exit.
	 */
	pthread_exit(0);
	pthread_cleanup_pop(1);

	return 0;
}

void
thread_kill(thread_t *thread)
{
	int err;

	assert(thread != thread_current());

	err = pthread_cancel(thread->id);
	assert(err == 0);
}

void
thread_exit(void)
{
	/* Just exit, the actual work is done in the cleanup handler */
	pthread_exit(0);
}

/*
 * Wake any thread waiting for this one to finish and wait for it to
 * acknowledge.
 */
static void
thread_done(thread_t *thread)
{
	threadpool_t *tpool = thread->pool;

	if (thread->chan) {
		channel_thread_exit(thread->chan);
		thread->chan = 0;
	}

	threadpool_lock(tpool);
	if (thread->flags & THREAD_WAITEDON) {
#ifdef VERBOSE
printf("TP%p: thread %p finishing\n", tpool, thread);
#endif
		thread->flags &= ~THREAD_WAITEDON;
		icond_signal(&thread->waitedon);
		thread->flags |= THREAD_DONE;
		while ((thread->flags & THREAD_DONE) != 0)
			icond_wait(&thread->waiting, &tpool->lock);
#ifdef VERBOSE
printf("TP%p: thread %p finished\n", tpool, thread);
#endif
	}
	assert((thread->flags & (THREAD_WAITEDON|THREAD_DONE)) == 0);
	threadpool_unlock(tpool);
}

static void
thread_cleanup(void *arg)
{
	thread_t *thread = arg;
	thread_t **pth, *th;
	threadpool_t *tpool;

	assert(thread == thread_current());

	tpool = thread->pool;
	threadpool_lock(tpool);
#ifdef VERBOSE
printf("TP%p: cleanup thread %p\n", tpool, thread);
#endif
	for (pth = &tpool->threads; (th = *pth) != 0; pth = &th->next)
		if (th == thread)
			break;
	assert(th != 0);
	*pth = thread->next;

	tpool->curthreads--;
	/*
	 * Can't cleanup the threadpool in the waiting case since
	 * the waiter will need to reacquire its lock in thread_run.
	 */
	if (thread->flags & THREAD_WAITEDON) {
		thread->flags &= ~THREAD_WAITEDON;
		icond_signal(&thread->waitedon);
		threadpool_unlock(tpool);
	} else {
		ilock_deinit(&thread->lock);
		icond_deinit(&thread->waitedon);
		mem_free(thread, sizeof *thread);
		if (tpool->curthreads == 0)
			threadpool_cleanup(tpool);
		else
			threadpool_unlock(tpool);
	}
}

void
thread_yield(void)
{
	sched_yield();
}

/*
 * XXX hack, uninterruptable (though cancelable) sleep implementation.
 * XXX assumes that an_timespec is same as timespec
 *
 * I'm not convinced thread_sleep is all that useful or I would spend some
 * time doing it right.
 */
void
thread_sleep(an_timespec_t *delay)
{
	int err;
	an_timespec_t wakeup;

	time_future(&wakeup, delay);

	err = pthread_mutex_lock(&sleepmutex);
	assert(err == 0);

	do {
		pthread_cleanup_push((void (*)(void *))pthread_mutex_unlock,
				     &sleepmutex);
		err = pthread_cond_timedwait(&sleepcond, &sleepmutex, &wakeup);
		pthread_cleanup_pop(0);
	} while (err == 0);
	assert(err == EINVAL || err == ETIMEDOUT);

	err = pthread_mutex_unlock(&sleepmutex);
	assert(err == 0);
}

void *
thread_getdata(void)
{
	return pthread_getspecific(nodeos_threaddata);
}

void
thread_setdata(void *data)
{
	int err;

	err = pthread_setspecific(nodeos_threaddata, data);
	assert(err == 0);
}

int
thread_getprio(thread_t *thread)
{
	return thread->nprio;
}

int
thread_setprio(thread_t *thread, int nprio)
{
	int err;

	err = os_sched_setprio(thread, nprio);
	if (err == 0)
		thread->nprio = nprio;

	return err;
}


/*
 * Thread "interrupt" support.
 *
 * With pthreads, signals are the only way we can do this.
 * Pthread cancelation implies the death of the canceled thread,
 * NodeOS interrupts do not.
 */
#include <signal.h>

#define SIG_THREADINTR	SIGUSR1 /* XXX */
sigset_t thread_intrmask;

static void thread_intr_handler(int sig, siginfo_t *si, void *sc);

static void
thread_intr_init(void)
{
	struct sigaction act;

	sigemptyset(&thread_intrmask);
	sigaddset(&thread_intrmask, SIG_THREADINTR);

	/*
	 * Catch relevant signals.
	 */
#ifdef SA_SIGINFO
	act.sa_flags = SA_SIGINFO;
	act.sa_sigaction = thread_intr_handler;
#else
	act.sa_handler = thread_intr_handler;
#endif
	sigemptyset(&act.sa_mask);

	sigaction(SIG_THREADINTR, &act, 0);
	sigaction(SIGILL, &act, 0);
#ifdef SIGEMT
	sigaction(SIGEMT, &act, 0);
#endif
	sigaction(SIGFPE, &act, 0);
	sigaction(SIGBUS, &act, 0);
	sigaction(SIGSEGV, &act, 0);
}

void
thread_interrupt(thread_t *thread)
{
	int err;
	cond_t *cond;

	thread_lock(thread);

	/*
	 * Send a signal to said thread
	 */
	err = pthread_kill(thread->id, SIG_THREADINTR);
	assert(err == 0);

	/*
	 * If thread is in a cond wait, wake it up.
	 * This gets really strange because we must acquire the cond
	 * lock before the thread lock (because of how thread_condwait
	 * works).
	 */
 again:
	cond = thread->cond;
	if (cond) {
		/*
		 * Cond lock is held.  Release the thread lock and wait
		 * for the cond.  Then reaquire the thread lock and see
		 * if it is still waiting on the same cond.  If not,
		 * start over.
		 */
		if (cond_trylock(cond)) {
			thread_unlock(thread);
			cond_lock(cond);
			thread_lock(thread);
			if (thread->cond != cond) {
				cond_unlock(cond);
				goto again;
			}
		}
		icond_broadcast(&cond->cond);
		cond_unlock(cond);
	}

	thread_unlock(thread);
}

void
thread_block_interrupt(void)
{
	thread_current()->flags &= ~THREAD_INTSON;
}

void
thread_unblock_interrupt(void)
{
	thread_current()->flags |= THREAD_INTSON;
}

void
thread_sethandler(thread_handler_t handler)
{
	thread_current()->handler = handler;
}

/*
 * Signal handler for all exceptions and NodeOS thread interrupts
 */
static void
thread_intr_handler(int sig, siginfo_t *si, void *sc)
{
	thread_t *thread = thread_current();
	int intson, type;

	if (thread == 0 || (thread->flags & THREAD_INKERNEL)) {
		os_dump_state(sc);
		panic("NodeOS: got signal %d\n", sig);
	}

	NODEOS_ENTER;

	/*
	 * No handler for thread, just exit
	 */
	 if (thread->handler == 0) {
		 printf("thread_intr_handler: thread has no handler:\n");
		 os_dump_state(sc);
		 pthread_cancel(thread->id);
		 return;
	 }

	/*
	 * Otherwise, call the handler with interrupts disabled
	 */
	intson = (thread->flags & THREAD_INTSON);
	thread->flags &= ~THREAD_INTSON;
	if (sig == SIG_THREADINTR) {
		assert(intson);
		type = AN_THREAD_INTR;
	} else
		type = AN_THREAD_TRAP;

	NODEOS_LEAVE;
	thread->handler(type, sc);
	NODEOS_ENTER;

	/* XXX NODEOS_LEAVE sans signal code */
	thread->flags |= intson;
	thread->flags &= ~THREAD_INKERNEL;

	NODEOS_LEAVE;
}

/*
 * Block a thread on the specified condition variable.
 * The thread may be awakened either by the condition being signaled through
 * an API call or because of thread_interrupt.  We do not differentiate the
 * two cases here.
 *
 * Called with thread and condition locked.
 */
void
thread_condwait(cond_t *cond)
{
	thread_t *thread = thread_current();
	assert(thread->cond == 0);

	thread->cond = cond;
	icond_wait(&cond->cond, &cond->lock);
	thread->cond = 0;
}
