/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 *
 */
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <setjmp.h>
#include <assert.h>
#include <nodeos/impl.h>

#include "threadpool.h"
#include "thread.h"
#include "flow.h"
#include "channel.h"
#include "mem.h"
#include "nodeos.h"

#undef VERBOSE

static int threadpool_thread_dowork(threadpool_t *tpool, channel_t *chan,
				    thread_func_t func, void *farg1,
				    void *farg2, int inkern, int canqueue,
				    thread_t **out_thread);
#ifdef THREADPOOL_STATS
static void threadpool_stats(threadpool_t *tpool);
#endif

int
threadpool_create(struct flow *flow, struct chan *chan, cpuspec_t *cspec,
		  threadpool_t **out_tpool)
{
	threadpool_t *tpool;
	int nthreads;
	int err;

	/* Verify the thread params with the underlying implementation */
	err = thread_checkattrs(cspec);
	if (err)
		return err;

	assert(flow->mpool);
	tpool = mempool_alloc(flow->mpool, sizeof *tpool);
	if (tpool == 0)
		return ENOMEM;

	memset(tpool, 0, sizeof *tpool);
	tpool->flow = flow;
	tpool->chan = chan;
	tpool->params = *cspec;
	icond_init(&tpool->waiting);
	ilock_init(&tpool->lock);

	/*
	 * XXX for now preallocate up to 4 threads
	 */
	nthreads = 4;
	if (nthreads > tpool->params.max)
		nthreads = tpool->params.max;
	threadpool_lock(tpool);
	threadpool_thread_setreserve(tpool, nthreads, 1);
	threadpool_unlock(tpool);

	*out_tpool = tpool;
	return 0;
}

void
threadpool_destroy(threadpool_t *tpool)
{
	thread_t *th, *me;

	me = thread_current();
	threadpool_lock(tpool);
	tpool->flags |= TPOOL_DYING;
	threadpool_thread_setreserve(tpool, 0, 0);
	for (th = tpool->threads; th != 0; th = th->next)
		if (th != me)
			thread_kill(th);

	if (tpool->curthreads == 0)
		threadpool_cleanup(tpool);
	else
		threadpool_unlock(tpool);
}

#if 0
/*
 * Similar to threadpool destroy, but kill only threads for a particular
 * channel (and don't destroy the pool!)
 *
 * Channel should be locked.
 */
void
threadpool_kill(threadpool_t *tpool, struct channel *chan)
{
	thread_t *th, *me;

	me = thread_current();
	threadpool_lock(tpool);
	for (th = tpool->threads; th != 0; th = th->next)
		if (th != me && th->chan == chan)
			thread_kill(th);
	threadpool_unlock(tpool);
}
#endif

void
threadpool_cleanup(threadpool_t *tpool)
{
	int flags;

	assert(threadpool_islocked(tpool));
	assert(tpool->curthreads == 0);

	flags = tpool->flags;
#ifdef THREADPOOL_STATS
	threadpool_stats(tpool);
#endif
	threadpool_unlock(tpool);

	if (flags & TPOOL_DYING) {
		assert(tpool->flow);
		icond_deinit(&tpool->waiting);
		ilock_deinit(&tpool->lock);
		mempool_free(tpool->flow->mpool, tpool, sizeof *tpool);
	}
}

/*
 * Schedule the function to be run in a thread from the indicated pool.
 * If no threads are currently available, the request will be queued.
 * Returns zero on success, an error code otherwise.
 */
int
threadpool_thread_schedule(threadpool_t *tpool, channel_t *chan, int inkern,
			   thread_func_t func, void *farg1, void *farg2)
{
	int err;

	threadpool_lock(tpool);
	err = threadpool_thread_dowork(tpool, chan, func, farg1, farg2,
				       inkern, 1, 0);
	threadpool_unlock(tpool);

	return err;
}

/*
 * Start a thread from the given pool running the indicated function.
 * If there are no threads available and the pool is at the thread limit,
 * returns an error.  If the pool is under limit but there are no cached
 * threads available, the call is delayed briefly while a new thread is started.
 * That is, upon return from this call, an actual thread has been commited
 * to run the function.  If out_threadp is non-zero, the call returns a
 * pointer to the new thread.  Returns zero if a thread was successfully
 * started, an error otherwise.
 */
int
threadpool_thread_start(threadpool_t *tpool, int inkern,
			thread_func_t func, void *farg1, void *farg2,
			thread_t **out_thread)
{
	int err;

	threadpool_lock(tpool);
	err = threadpool_thread_dowork(tpool, 0, func, farg1, farg2, inkern, 0,
				      out_thread);
	threadpool_unlock(tpool);

	return err;
}

/*
 * Start a thread from the given pool and wait til it terminates or the
 * timeout expires.  If the timeout expires, the thread is killed and a
 * distinguished error code is returned.  Returns zero on success, an error
 * code otherwise.
 *
 * XXX implement timeout
 */
int
threadpool_thread_run(threadpool_t *tpool, int inkern,
		      thread_func_t func, void *farg1, void *farg2,
		      an_timespec_t *timo)
{
	thread_t *thread;
	int err;

	threadpool_lock(tpool);

	/*
	 * Fire off the thread
	 */
	err = threadpool_thread_dowork(tpool, 0, func, farg1, farg2, inkern, 0,
				       &thread);
	if (err) {
		threadpool_unlock(tpool);
		return err;
	}

	/*
	 * Wait for it to finish
	 */
	thread->flags |= THREAD_WAITEDON;
	while ((thread->flags & THREAD_WAITEDON) != 0)
		icond_wait(&thread->waitedon, &tpool->lock);

	/*
	 * Allow it to go service another request or terminate.
	 */
	assert((thread->flags & THREAD_DONE) != 0);
	thread->flags &= ~THREAD_DONE;
	icond_signal(&thread->waiting);

	threadpool_unlock(tpool);
	return 0;
}

/*
 * Start a thread in the given pool to run the indicated function.
 * If inkern is non-zero, the thread is run in the NodeOS context.
 * If canqueue is non-zero then, if no threads are available, the record
 * is put on the threadpool workqueue to be handled later.  If out_thread
 * is non-zero, the thread chosen to process the request is returned.
 */
static int
threadpool_thread_dowork(threadpool_t *tpool, channel_t *chan,
			 thread_func_t func, void *farg1, void *farg2,
			 int inkern, int canqueue, thread_t **out_thread)
{
	struct threadwork *w;
	thread_t *thread;
	int err;

	/*
	 * Create a work record to give to a thread or queue.
	 */
	assert(tpool->flow);
	w = mempool_alloc(tpool->flow->mpool, sizeof *w);
	if (w == 0)
		return ENOMEM;
	w->inkern = inkern;
	w->chan = chan;
	w->func = func;
	w->farg[0] = farg1;
	w->farg[1] = farg2;

	assert(threadpool_islocked(tpool));

#ifdef VERBOSE
printf("TP%p: starting %p(%p): ", tpool, w->func, w->farg);
#endif
#ifdef THREADPOOL_STATS
	tpool->stats.starts++; 
#endif

	/*
	 * If there is a free pool thread available now, give it the work
	 * record and wake it up.
	 */
	if (tpool->freethreads != 0) {
		assert(tpool->freelist != 0);
		thread = tpool->freelist;
		tpool->freelist = thread->freenext;
		tpool->freethreads--;
		thread->freenext = 0;
		thread->work = w;
		icond_signal(&thread->waiting);
		if (out_thread)
			*out_thread = thread;
#ifdef VERBOSE
printf("woke cached thread %p\n", thread);
#endif
		return 0;
	}

	/*
	 * If we are not at the thread limit for the pool, start up a new
	 * thread and give it the work.
	 */
	err = thread_spawn(tpool, &thread);
	if (err == 0) {
		thread->work = w;
		if (out_thread)
			*out_thread = thread;
#ifdef VERBOSE
printf("spawned new thread %p\n", thread);
#endif
#ifdef THREADPOOL_STATS
		tpool->stats.creates++;
#endif
		return 0;
	}

	/*
	 * No threads are available.  If it is ok to queue the request,
	 * put it on the end of the work list for the pool.  Otherwise
	 * we return an error.
	 */
	if (canqueue) {
		if (tpool->worktail == 0) {
			assert(tpool->workhead == 0);
			tpool->workhead = tpool->worktail = w;
		} else {
			assert(tpool->workhead != 0);
			tpool->worktail->next = w;
			tpool->worktail = w;
		}
		tpool->worklen++;
#ifdef THREADPOOL_STATS
		if (tpool->worklen > tpool->stats.maxqueued)
			tpool->stats.maxqueued = tpool->worklen;
#endif
		if (out_thread)
			*out_thread = 0;
#ifdef VERBOSE
printf("queued work\n");
#endif
#ifdef THREADPOOL_STATS
		tpool->stats.queued++;
#endif
		return 0;
	}

	mempool_free(tpool->flow->mpool, w, sizeof *w);
#ifdef VERBOSE
printf("failed\n");
#endif
#ifdef THREADPOOL_STATS
	tpool->stats.failed++;
#endif
	return err;
}

/*
 * Ensure there are the indicated number of threads in a pool's freelist now
 * and make that value the target freelist size.  Pool should be locked on
 * entry and exit.
 */
int
threadpool_thread_setreserve(threadpool_t *tpool, int nthreads, int wait)
{
	int delta, err;
	thread_t *thread;

	assert(threadpool_islocked(tpool));

	delta = nthreads - tpool->freethreads;

#ifdef VERBOSE
printf("TP%p: setreserve %d\n", tpool, nthreads);
#endif
	/*
	 * Not enough free threads currently, make more.
	 *
	 * XXX note that we might fail after creating some threads.
	 * In this case we return an error without updating the reserve count.
	 * The extra threads we did spawn will die when they start up since
	 * we haven't updated the reserve.
	 */
	while (delta > 0) {
		err = thread_spawn(tpool, 0);
#ifdef VERBOSE
printf("setreserve create thread: err=%d\n", err);
#endif
		if (err)
			return err;
		delta--;
	}

	/*
	 * Too many threads, pull off the appropriate number
	 * and wake them up so they will exit.
	 */
	while (delta < 0) {
		tpool->freethreads--;
		thread = tpool->freelist;
		assert(thread->work == 0);
		tpool->freelist = thread->freenext;
		icond_signal(&thread->waiting);
		delta++;
#ifdef VERBOSE
printf("setreserve wake thread %p\n", thread);
#endif
	}
	tpool->reserve = nthreads;

	/*
	 * Wait for the thread count to hit the desired value.
	 * The threadpool had better be idle or this wait might not return.
	 *
	 * XXX only works for threads being created.
	 */
	if (wait) {
		while (tpool->freethreads != tpool->reserve) {
			tpool->flags |= TPOOL_WAITING;
			icond_wait(&tpool->waiting, &tpool->lock);
		}
	}

	return 0;
}

/*
 * Return a thread to the threadpool reserve
 */
void
threadpool_thread_exit(threadpool_t *tpool)
{
	thread_t *thread = thread_current();
	assert(tpool == thread->pool);

	siglongjmp(thread->jmpbuf, 1);
}

#ifdef THREADPOOL_STATS
#include <stdio.h>

static void
threadpool_stats(threadpool_t *tpool)
{
	int cached;

	cached = tpool->stats.starts -
		(tpool->stats.creates+tpool->stats.queued+tpool->stats.failed);
	printf("TP%p: %d thread starts\n", tpool, tpool->stats.starts);
	printf("\t(%d cached, %d created, %d queued, %d failed)\n",
	       cached, tpool->stats.creates, tpool->stats.queued,
	       tpool->stats.failed);
	printf("  Max: %d active threads, %d free threads, %d queued\n",
	       tpool->stats.maxactive, tpool->stats.maxcached,
	       tpool->stats.maxqueued);
}
#endif
